
from openmdao.api import ExplicitComponent, Group, IndepVarComp
import numpy as np
import wisdem.pyframe3dd.frame3dd as frame3dd
from wisdem.commonse.utilities import nodal2sectional

from wisdem.commonse import gravity, eps, Tube, NFREQ
import wisdem.commonse.UtilizationSupplement as util
import wisdem.commonse.manufacturing as manufacture
from wisdem.commonse.WindWaveDrag import AeroHydroLoads, CylinderWindDrag, CylinderWaveDrag
from wisdem.commonse.environment import WaveBase, PowerWind
from wisdem.commonse.vertical_cylinder import CylinderDiscretization, CylinderMass
from .map_mooring import NLINES_MAX


def find_nearest(array,value):
    return (np.abs(array-value)).argmin() 

        
def ghostNodes(x1, x2, r1, r2):
    dx = x2 - x1
    L = np.sqrt( np.sum( dx**2 ) )
    dr1 = (      r1/L) * dx + x1
    dr2 = (1.0 - r2/L) * dx + x1
    return dr1, dr2

class FloatingFrame(ExplicitComponent):
    """
    OpenMDAO Component class for semisubmersible pontoon / truss structure for floating offshore wind turbines.
    Should be tightly coupled with Semi and Mooring classes for full system representation.
    """

    def initialize(self):
        self.options.declare('nFull')
        self.options.declare('nFullTow')
        
    def setup(self):
        nFull    = self.options['nFull']
        nFullTow = self.options['nFullTow']

        # Keep Frame3DD data object for easy testing and debugging
        self.myframe = None
        
        # Environment
        self.add_input('water_density', val=0.0, units='kg/m**3', desc='density of water')

        # Material properties
        self.add_input('material_density', val=0., units='kg/m**3', desc='density of material')
        self.add_input('E', val=0.0, units='Pa', desc='Modulus of elasticity (Youngs) of material')
        self.add_input('G', val=0.0, units='Pa', desc='Shear modulus of material')
        self.add_input('yield_stress', val=0.0, units='Pa', desc='yield stress of material')
        self.add_input('Hs', val=0.0, units='m', desc='wave significant height')

        # Base column
        self.add_input('main_z_full', val=np.zeros((nFull,)), units='m', desc='z-coordinates of section nodes (length = nsection+1)')
        self.add_input('main_d_full', val=np.zeros((nFull,)), units='m', desc='outer radius at each section node bottom to top (length = nsection + 1)')
        self.add_input('main_t_full', val=np.zeros((nFull-1,)), units='m', desc='shell wall thickness at each section node bottom to top (length = nsection + 1)')
        self.add_input('main_mass', val=np.zeros((nFull-1,)), units='kg', desc='mass of main column by section')
        self.add_input('main_buckling_length', val=np.zeros((nFull-1,)), units='m', desc='distance between ring stiffeners')
        self.add_input('main_displaced_volume', val=np.zeros((nFull-1,)), units='m**3', desc='column volume of water displaced by section')
        self.add_input('main_hydrostatic_force', val=np.zeros((nFull-1,)), units='N', desc='Net z-force of hydrostatic pressure by section')
        self.add_input('main_center_of_buoyancy', val=0.0, units='m', desc='z-position of center of column buoyancy force')
        self.add_input('main_center_of_mass', val=0.0, units='m', desc='z-position of center of column mass')
        self.add_input('main_Px', np.zeros(nFull), units='N/m', desc='force per unit length in x-direction on main')
        self.add_input('main_Py', np.zeros(nFull), units='N/m', desc='force per unit length in y-direction on main')
        self.add_input('main_Pz', np.zeros(nFull), units='N/m', desc='force per unit length in z-direction on main')
        self.add_input('main_qdyn', np.zeros(nFull), units='N/m**2', desc='dynamic pressure on main')

        self.add_input('main_pontoon_attach_upper', val=0.0, desc='Fraction of main column for upper truss attachment on main column')
        self.add_input('main_pontoon_attach_lower', val=0.0, desc='Fraction of main column lower truss attachment on main column')

        # offset columns
        self.add_input('offset_z_full', val=np.zeros((nFull,)), units='m', desc='z-coordinates of section nodes (length = nsection+1)')
        self.add_input('offset_d_full', val=np.zeros((nFull,)), units='m', desc='outer radius at each section node bottom to top (length = nsection + 1)')
        self.add_input('offset_t_full', val=np.zeros((nFull-1,)), units='m', desc='shell wall thickness at each section node bottom to top (length = nsection + 1)')
        self.add_input('offset_mass', val=np.zeros((nFull-1,)), units='kg', desc='mass of offset column by section')
        self.add_input('offset_buckling_length', val=np.zeros((nFull-1,)), units='m', desc='distance between ring stiffeners')
        self.add_input('offset_displaced_volume', val=np.zeros((nFull-1,)), units='m**3', desc='column volume of water displaced by section')
        self.add_input('offset_hydrostatic_force', val=np.zeros((nFull-1,)), units='N', desc='Net z-force of hydrostatic pressure by section')
        self.add_input('offset_center_of_buoyancy', val=0.0, units='m', desc='z-position of center of column buoyancy force')
        self.add_input('offset_center_of_mass', val=0.0, units='m', desc='z-position of center of column mass')
        self.add_input('offset_Px', np.zeros(nFull), units='N/m', desc='force per unit length in x-direction on offset')
        self.add_input('offset_Py', np.zeros(nFull), units='N/m', desc='force per unit length in y-direction on offset')
        self.add_input('offset_Pz', np.zeros(nFull), units='N/m', desc='force per unit length in z-direction on offset')
        self.add_input('offset_qdyn', np.zeros(nFull), units='N/m**2', desc='dynamic pressure on offset')

        # Tower
        self.add_input('tower_z_full', val=np.zeros((nFullTow,)), units='m', desc='z-coordinates of section nodes (length = nsection+1)')
        self.add_input('tower_d_full', val=np.zeros((nFullTow,)), units='m', desc='outer radius at each section node bottom to top (length = nsection + 1)')
        self.add_input('tower_t_full', val=np.zeros((nFullTow-1,)), units='m', desc='shell wall thickness at each section node bottom to top (length = nsection + 1)')
        self.add_input('tower_mass_section', val=np.zeros((nFullTow-1,)), units='kg', desc='mass of tower column by section')
        self.add_input('tower_buckling_length', 0.0, units='m', desc='buckling length')
        self.add_input('tower_center_of_mass', val=0.0, units='m', desc='z-position of center of tower mass')
        self.add_input('tower_Px', np.zeros(nFullTow), units='N/m', desc='force per unit length in x-direction on tower')
        self.add_input('tower_Py', np.zeros(nFullTow), units='N/m', desc='force per unit length in y-direction on tower')
        self.add_input('tower_Pz', np.zeros(nFullTow), units='N/m', desc='force per unit length in z-direction on tower')
        self.add_input('tower_qdyn', np.zeros(nFullTow), units='N/m**2', desc='dynamic pressure on tower')
        
        # Semi geometry
        self.add_input('radius_to_offset_column', val=0.0, units='m',desc='Distance from main column centerpoint to offset column centerpoint')
        self.add_input('number_of_offset_columns', val=3, desc='Number of offset columns evenly spaced around main column')

        # Pontoon properties
        self.add_input('pontoon_outer_diameter', val=0.0, units='m',desc='Outer radius of tubular pontoon that connects offset or main columns')
        self.add_input('pontoon_wall_thickness', val=0.0, units='m',desc='Inner radius of tubular pontoon that connects offset or main columns')
        self.add_discrete_input('cross_attachment_pontoons', val=True, desc='Inclusion of pontoons that connect the bottom of the central main to the tops of the outer offset columns')
        self.add_discrete_input('lower_attachment_pontoons', val=True, desc='Inclusion of pontoons that connect the central main to the outer offset columns at their bottoms')
        self.add_discrete_input('upper_attachment_pontoons', val=True, desc='Inclusion of pontoons that connect the central main to the outer offset columns at their tops')
        self.add_discrete_input('lower_ring_pontoons', val=True, desc='Inclusion of pontoons that ring around outer offset columns at their bottoms')
        self.add_discrete_input('upper_ring_pontoons', val=True, desc='Inclusion of pontoons that ring around outer offset columns at their tops')
        self.add_discrete_input('outer_cross_pontoons', val=True, desc='Inclusion of pontoons that ring around outer offset columns at their tops')
        
        # Turbine parameters
        self.add_input('rna_mass', val=0.0, units='kg', desc='mass of tower')
        self.add_input('rna_cg', val=np.zeros(3), units='m', desc='Location of RNA center of mass relative to tower top')
        self.add_input('rna_force', val=np.zeros(3), units='N', desc='Force in xyz-direction on turbine')
        self.add_input('rna_moment', val=np.zeros(3), units='N*m', desc='Moments about turbine main')
        self.add_input('rna_I', val=np.zeros(6), units='kg*m**2', desc='Moments about turbine main')

        # Mooting parameters for loading
        self.add_input('number_of_mooring_connections', val=3, desc='number of mooring connections on vessel')
        self.add_input('mooring_lines_per_connection', val=1, desc='number of mooring lines per connection')
        self.add_input('mooring_neutral_load', val=np.zeros((NLINES_MAX,3)), units='N', desc='z-force of mooring lines on structure')
        self.add_input('mooring_stiffness', val=np.zeros((6,6)), units='N/m', desc='Linearized stiffness matrix of mooring system at neutral (no offset) conditions.')
        self.add_input('mooring_moments_of_inertia', val=np.zeros(6), units='kg*m**2', desc='mass moment of inertia of mooring system about fairlead-centerline point [xx yy zz xy xz yz]')
        self.add_input('fairlead', val=0.0, units='m', desc='Depth below water for mooring line attachment')
        self.add_input('fairlead_radius', val=0.0, units='m',desc='Radius from center of structure to fairlead connection points')
        self.add_input('fairlead_support_outer_diameter', val=0.0, units='m',desc='fairlead support outer diameter')
        self.add_input('fairlead_support_wall_thickness', val=0.0, units='m',desc='fairlead support wall thickness')
        
        # safety factors
        self.add_input('gamma_f', 0.0, desc='safety factor on loads')
        self.add_input('gamma_m', 0.0, desc='safety factor on materials')
        self.add_input('gamma_n', 0.0, desc='safety factor on consequence of failure')
        self.add_input('gamma_b', 0.0, desc='buckling safety factor')
        self.add_input('gamma_fatigue', 0.0, desc='total safety factor for fatigue')

        # Manufacturing
        self.add_input('connection_ratio_max', val=0.0, desc='Maximum ratio of pontoon outer diameter to main/offset outer diameter')
        
        # Costing
        self.add_input('material_cost_rate', 0.0, units='USD/kg', desc='Raw material cost rate: steel $1.1/kg, aluminum $3.5/kg')
        self.add_input('labor_cost_rate', 0.0, units='USD/min', desc='Labor cost rate')
        self.add_input('painting_cost_rate', 0.0, units='USD/m/m', desc='Painting / surface finishing cost rate')
        
        # Outputs
        self.add_output('pontoon_wave_height_depth_margin', val=np.zeros(2), units='m', desc='Distance between attachment point of pontoons and wave crest- both above and below waterline')
        self.add_output('pontoon_cost', val=0.0, units='USD', desc='Cost of pontoon elements and connecting truss')
        self.add_output('pontoon_cost_rate', val=0.0, units='USD/t', desc='Cost rate of finished pontoon and truss')
        self.add_output('pontoon_mass', val=0.0, units='kg', desc='Mass of pontoon elements and connecting truss')
        self.add_output('pontoon_displacement', val=0.0, units='m**3', desc='Buoyancy force of submerged pontoon elements')
        self.add_output('pontoon_center_of_buoyancy', val=0.0, units='m', desc='z-position of center of pontoon buoyancy force')
        self.add_output('pontoon_center_of_mass', val=0.0, units='m', desc='z-position of center of pontoon mass')

        self.add_output('top_deflection', 0.0, units='m', desc='Deflection of tower top in yaw-aligned +x direction')
        self.add_output('pontoon_stress', val=np.zeros((70,)), desc='Utilization (<1) of von Mises stress by yield stress and safety factor for all pontoon elements')

        self.add_output('main_stress', np.zeros(nFull-1), desc='Von Mises stress utilization along main column at specified locations. Incudes safety factor.')
        self.add_output('main_stress:axial', np.zeros(nFull-1), desc='Axial stress along main column at specified locations.')
        self.add_output('main_stress:shear', np.zeros(nFull-1), desc='Shear stress along main column at specified locations.')
        self.add_output('main_stress:hoop', np.zeros(nFull-1), desc='Hoop stress along main column at specified locations.')
        self.add_output('main_stress:hoopStiffen', np.zeros(nFull-1), desc='Hoop stress along main column at specified locations.')
        self.add_output('main_shell_buckling', np.zeros(nFull-1), desc='Shell buckling constraint. Should be < 1 for feasibility. Includes safety factors')
        self.add_output('main_global_buckling', np.zeros(nFull-1), desc='Global buckling constraint. Should be < 1 for feasibility. Includes safety factors')

        self.add_output('offset_stress', np.zeros(nFull-1), desc='Von Mises stress utilization along offset column at specified locations. Incudes safety factor.')
        self.add_output('offset_stress:axial', np.zeros(nFull-1), desc='Axial stress along offset column at specified locations.')
        self.add_output('offset_stress:shear', np.zeros(nFull-1), desc='Shear stress along offset column at specified locations.')
        self.add_output('offset_stress:hoop', np.zeros(nFull-1), desc='Hoop stress along offset column at specified locations.')
        self.add_output('offset_stress:hoopStiffen', np.zeros(nFull-1), desc='Hoop stress along offset column at specified locations.')
        self.add_output('offset_shell_buckling', np.zeros(nFull-1), desc='Shell buckling constraint. Should be < 1 for feasibility. Includes safety factors')
        self.add_output('offset_global_buckling', np.zeros(nFull-1), desc='Global buckling constraint. Should be < 1 for feasibility. Includes safety factors')

        self.add_output('tower_stress', np.zeros(nFullTow-1), desc='Von Mises stress utilization along tower at specified locations.  incudes safety factor.')
        self.add_output('tower_stress:axial', np.zeros(nFullTow-1), desc='Axial stress along tower column at specified locations.')
        self.add_output('tower_stress:shear', np.zeros(nFullTow-1), desc='Shear stress along tower column at specified locations.')
        self.add_output('tower_stress:hoop', np.zeros(nFullTow-1), desc='Hoop stress along tower column at specified locations.')
        self.add_output('tower_stress:hoopStiffen', np.zeros(nFullTow-1), desc='Hoop stress along tower column at specified locations.')
        self.add_output('tower_shell_buckling', np.zeros(nFullTow-1), desc='Shell buckling constraint.  Should be < 1 for feasibility.  Includes safety factors')
        self.add_output('tower_global_buckling', np.zeros(nFullTow-1), desc='Global buckling constraint.  Should be < 1 for feasibility.  Includes safety factors')

        self.add_discrete_output('plot_matrix', val=np.array([]), desc='Ratio of shear stress to yield stress for all pontoon elements')
        self.add_output('main_connection_ratio', val=np.zeros((nFull,)), desc='Ratio of pontoon outer diameter to main outer diameter')
        self.add_output('offset_connection_ratio', val=np.zeros((nFull,)), desc='Ratio of pontoon outer diameter to main outer diameter')

        self.add_output('structural_frequencies', np.zeros(NFREQ), units='Hz', desc='First six natural frequencies')
        self.add_output('substructure_mass', val=0.0, units='kg', desc='Mass of substructure elements and connecting truss')
        self.add_output('structural_mass', val=0.0, units='kg', desc='Mass of whole turbine except for mooring lines')
        self.add_output('total_displacement', val=0.0, units='m**3', desc='Total volume of water displaced by floating turbine (except for mooring lines)')
        self.add_output('z_center_of_buoyancy', val=0.0, units='m', desc='z-position of center of buoyancy of whole turbine')
        self.add_output('substructure_center_of_mass', val=np.zeros(3), units='m', desc='xyz-position of center of gravity of substructure only')
        self.add_output('structure_center_of_mass', val=np.zeros(3), units='m', desc='xyz-position of center of gravity of whole turbine')
        self.add_output('total_force', val=np.zeros(3), units='N', desc='Net forces on turbine')
        self.add_output('total_moment', val=np.zeros(3), units='N*m', desc='Moments on whole turbine')
        
        # Derivatives
        self.declare_partials('*', '*', method='fd', form='central', step=1e-6)
         
    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
        
        # Unpack variables
        ncolumn         = int(inputs['number_of_offset_columns'])
        crossAttachFlag = discrete_inputs['cross_attachment_pontoons']
        lowerAttachFlag = discrete_inputs['lower_attachment_pontoons']
        upperAttachFlag = discrete_inputs['upper_attachment_pontoons']
        lowerRingFlag   = discrete_inputs['lower_ring_pontoons']
        upperRingFlag   = discrete_inputs['upper_ring_pontoons']
        outerCrossFlag  = discrete_inputs['outer_cross_pontoons']
        
        R_semi         = inputs['radius_to_offset_column'] if ncolumn>0 else 0.0
        R_od_pontoon   = 0.5*inputs['pontoon_outer_diameter']
        R_od_main      = 0.5*inputs['main_d_full']
        R_od_offset   = 0.5*inputs['offset_d_full']
        R_od_tower     = 0.5*inputs['tower_d_full']
        R_od_fairlead  = 0.5*inputs['fairlead_support_outer_diameter']

        t_wall_main     = inputs['main_t_full']
        t_wall_offset  = inputs['offset_t_full']
        t_wall_pontoon  = inputs['pontoon_wall_thickness']
        t_wall_tower    = inputs['tower_t_full']
        t_wall_fairlead = inputs['fairlead_support_wall_thickness']

        E              = inputs['E']
        G              = inputs['G']
        rho            = inputs['material_density']
        sigma_y        = inputs['yield_stress']
        
        z_main         = inputs['main_z_full']
        z_offset       = inputs['offset_z_full']
        z_tower        = inputs['tower_z_full']
        z_attach_upper = inputs['main_pontoon_attach_upper']*(z_main[-1] - z_main[0]) + z_main[0]
        z_attach_lower = inputs['main_pontoon_attach_lower']*(z_main[-1] - z_main[0]) + z_main[0]
        z_fairlead     = -inputs['fairlead']

        m_main         = inputs['main_mass']
        m_offset      = inputs['offset_mass']
        m_tower        = inputs['tower_mass_section']
        
        m_rna          = inputs['rna_mass']
        F_rna          = inputs['rna_force']
        M_rna          = inputs['rna_moment']
        I_rna          = inputs['rna_I']
        cg_rna         = inputs['rna_cg']
        
        rhoWater       = inputs['water_density']
        
        V_main         = inputs['main_displaced_volume']
        V_offset      = inputs['offset_displaced_volume']

        F_hydro_main    = inputs['main_hydrostatic_force']
        F_hydro_offset = inputs['offset_hydrostatic_force']

        z_cb_main      = inputs['main_center_of_buoyancy']
        z_cb_offset   = inputs['offset_center_of_buoyancy']
        
        cg_main        = np.r_[0.0, 0.0, inputs['main_center_of_mass']]
        cg_offset     = np.r_[0.0, 0.0, inputs['offset_center_of_mass']]
        cg_tower       = np.r_[0.0, 0.0, inputs['tower_center_of_mass']]
        
        n_connect      = int(inputs['number_of_mooring_connections'])
        n_lines        = int(inputs['mooring_lines_per_connection'])
        K_mooring      = np.diag( inputs['mooring_stiffness'] )
        I_mooring      = inputs['mooring_moments_of_inertia']
        F_mooring      = inputs['mooring_neutral_load']
        R_fairlead     = inputs['fairlead_radius']
        
        gamma_f        = inputs['gamma_f']
        gamma_m        = inputs['gamma_m']
        gamma_n        = inputs['gamma_n']
        gamma_b        = inputs['gamma_b']
        gamma_fatigue  = inputs['gamma_fatigue']

        # Quick ratio for unknowns
        outputs['main_connection_ratio']    = inputs['connection_ratio_max'] - R_od_pontoon/R_od_main
        outputs['offset_connection_ratio'] = inputs['connection_ratio_max'] - R_od_pontoon/R_od_offset
        outputs['pontoon_wave_height_depth_margin'] = np.abs(np.r_[z_attach_lower, z_attach_upper]) - np.abs(inputs['Hs'])

        
        # --- INPUT CHECKS -----
        # If something fails, we have to tell the optimizer this design is no good
        def bad_input():
            outputs['structural_frequencies'] = 1e30 * np.ones(NFREQ)
            outputs['top_deflection'] = 1e30
            outputs['substructure_mass']  = 1e30
            outputs['structural_mass']    = 1e30
            outputs['total_displacement'] = 1e30
            outputs['z_center_of_buoyancy'] = 0.0
            outputs['substructure_center_of_mass'] = 1e30 * np.ones(3)
            outputs['structure_center_of_mass'] = 1e30 * np.ones(3)
            outputs['total_force'] =  1e30 * np.ones(3)
            outputs['total_moment'] = 1e30 * np.ones(3)
            outputs['tower_stress'] = 1e30 * np.ones(m_tower.shape)
            outputs['tower_shell_buckling'] = 1e30 * np.ones(m_tower.shape)
            outputs['tower_global_buckling'] = 1e30 * np.ones(m_tower.shape)
            outputs['main_stress'] = 1e30 * np.ones(m_main.shape)
            outputs['main_shell_buckling'] = 1e30 * np.ones(m_main.shape)
            outputs['main_global_buckling'] = 1e30 * np.ones(m_main.shape)
            outputs['offset_stress'] = 1e30 * np.ones(m_offset.shape)
            outputs['offset_shell_buckling'] = 1e30 * np.ones(m_offset.shape)
            outputs['offset_global_buckling'] = 1e30 * np.ones(m_offset.shape)
            return
        
        # There is no truss if not offset columns
        if ncolumn == 0:
            crossAttachFlag = lowerAttachFlag = upperAttachFlag = False
            lowerRingFlag = upperRingFlag = outerCrossFlag  = False

        # Must have symmetry for the substructure to work out
        if ncolumn in [1, 2] or ncolumn > 7:
            bad_input()
            return

        # Must have symmetry in moorning loading too
        if (ncolumn > 0) and (n_connect > 0) and (ncolumn != n_connect):
            bad_input()
            return

        # If there are offset columns, must have attachment pontoons (only have ring pontoons doesn't make sense)
        if (ncolumn > 0) and (not crossAttachFlag) and (not lowerAttachFlag) and (not upperAttachFlag):
            bad_input()
            return

        # Must have lower ring if have cross braces
        if (ncolumn > 0) and outerCrossFlag and (not lowerRingFlag):
            bad_input()
            return

        # ---GEOMETRY---
        # Compute frustum angles
        angle_tower   = np.arctan( np.diff(R_od_tower)   / np.diff(z_tower)   )
        angle_main    = np.arctan( np.diff(R_od_main)    / np.diff(z_main)    )
        angle_offset = np.arctan( np.diff(R_od_offset) / np.diff(z_offset) )
        
        # ---NODES---
        # Add nodes for main column: Using 4 nodes/3 elements per section
        # Make sure there is a node at upper and lower attachment points
        mainBeginID = 0 + 1
        if ncolumn > 0:
            idx = find_nearest(z_main, z_attach_lower)
            z_main[idx] = z_attach_lower
            mainLowerID = idx + 1
            
            idx = find_nearest(z_main, z_attach_upper)
            z_main[idx] = z_attach_upper
            mainUpperID = idx + 1
        
        mainEndID = z_main.size
        freeboard = z_main[-1]

        fairleadID  = []
        # Need mooring attachment point if just running a spar
        if ncolumn == 0:
            idx = find_nearest(z_main, z_fairlead)
            z_main[idx] = z_fairlead
            fairleadID.append( idx + 1 )
        
        znode = np.copy( z_main )
        xnode = np.zeros(znode.shape)
        ynode = np.zeros(znode.shape)
        rnode = np.copy( R_od_main)

        towerBeginID = mainEndID
        myz = np.zeros(len(z_tower)-1)
        xnode = np.append(xnode, myz)
        ynode = np.append(ynode, myz)
        znode = np.append(znode, z_tower[1:] + freeboard )
        rnode = np.append(rnode, R_od_tower[1:])
        towerEndID = xnode.size

        # Create dummy node so that the tower isn't the last in a chain.
        # This avoids a Frame3DD bug
        dummyID = xnode.size + 1
        xnode = np.append(xnode, 0.0)
        ynode = np.append(ynode, 0.0)
        znode = np.append(znode, znode[-1]+1.0 )
        rnode = np.append(rnode, 0.0)
        
        # Get x and y positions of surrounding offset columns
        offsetLowerID = []
        offsetUpperID = []
        offsetx = R_semi * np.cos( np.linspace(0, 2*np.pi, ncolumn+1) )
        offsety = R_semi * np.sin( np.linspace(0, 2*np.pi, ncolumn+1) )
        offsetx = offsetx[:-1]
        offsety = offsety[:-1]

        # Add in offset column nodes around the circle, make sure there is a node at the fairlead
        idx = find_nearest(z_offset, z_fairlead)
        myones = np.ones(z_offset.shape)
        for k in range(ncolumn):
            offsetLowerID.append( xnode.size + 1 )
            fairleadID.append( xnode.size + idx + 1 )
            xnode = np.append(xnode, offsetx[k]*myones)
            ynode = np.append(ynode, offsety[k]*myones)
            znode = np.append(znode, z_offset )
            rnode = np.append(rnode, R_od_offset )
            offsetUpperID.append( xnode.size )

        # Add nodes where mooring lines attach, which may be offset from columns
        mooringx  = R_fairlead * np.cos( np.linspace(0, 2*np.pi, n_connect+1) )[:-1]
        mooringy  = R_fairlead * np.sin( np.linspace(0, 2*np.pi, n_connect+1) )[:-1]
        mooringID = xnode.size + 1 + np.arange(n_connect, dtype=np.int32)
        xnode     = np.append(xnode, mooringx)
        ynode     = np.append(ynode, mooringy)
        znode     = np.append(znode, z_fairlead*np.ones(n_connect) )
        rnode     = np.append(rnode, np.zeros(n_connect))
            
        # Add nodes midway around outer ring for cross bracing
        if outerCrossFlag and ncolumn > 0:
            crossx = 0.5*(offsetx + np.roll(offsetx,1))
            crossy = 0.5*(offsety + np.roll(offsety,1))

            crossOuterLowerID = xnode.size + np.arange(ncolumn) + 1
            crossOuterLowerID = crossOuterLowerID.tolist()
            xnode = np.append(xnode, crossx)
            ynode = np.append(ynode, crossy)
            znode = np.append(znode, z_offset[0]*np.ones(ncolumn))
            rnode = np.append(rnode, np.zeros(ncolumn))

            #crossOuterUpperID = xnode.size + np.arange(ncolumn) + 1
            #xnode = np.append(xnode, crossx)
            #ynode = np.append(ynode, crossy)
            #znode = np.append(znode, z_offset[-1]*np.ones(ncolumn))

        # Create matrix for easy referencing
        nodeMat = np.c_[xnode, ynode, znode]

        # To aid in wrap-around references
        if ncolumn > 0:
            offsetLowerID.append( offsetLowerID[0] )
            offsetUpperID.append( offsetUpperID[0] )
            if outerCrossFlag:
                crossOuterLowerID.append( crossOuterLowerID[0] )
        
        
        # ---ELEMENTS / EDGES---
        # To accurately capture pontoon length and stiffness, for each connection we create 2 additional nodes,
        # where the pontoon "line" intersects the main and offset shells.  Highly stiff "ghost" elements are created
        # from the column centerline to the shell.  These are not calculated for pontoon weight.
        # The actual pontoon only extends from shell boundary to shell boundary.
        N1  = np.array([], dtype=np.int32)
        N2  = np.array([], dtype=np.int32)
        gN1 = np.array([], dtype=np.int32)
        gN2 = np.array([], dtype=np.int32)
        
        # Lower connection from central main column to offset columns
        if lowerAttachFlag:
            lowerAttachEID = N1.size + 1
            for k in range(ncolumn):
                tempID1 = mainLowerID
                tempID2 = offsetLowerID[k]
                add1, add2 = ghostNodes(nodeMat[tempID1-1,:], nodeMat[tempID2-1,:], rnode[tempID1-1], rnode[tempID2-1])
                if ( (add1[-1] >= z_main[0]) and (add1[-1] <= z_main[-1]) ):
                    tempID1 = xnode.size + 1
                    xnode   = np.append(xnode, add1[0])
                    ynode   = np.append(ynode, add1[1])
                    znode   = np.append(znode, add1[2])
                    gN1     = np.append(gN1, mainLowerID)
                    gN2     = np.append(gN2, tempID1)
                if ( (add2[-1] >= z_offset[0]) and (add2[-1] <= z_offset[-1]) ):
                    tempID2 = xnode.size + 1
                    xnode   = np.append(xnode, add2[0])
                    ynode   = np.append(ynode, add2[1])
                    znode   = np.append(znode, add2[2])
                    gN1     = np.append(gN1, offsetLowerID[k])
                    gN2     = np.append(gN2, tempID2)
                # Pontoon connection
                N1 = np.append(N1, tempID1 )
                N2 = np.append(N2, tempID2 )
                
        # Upper connection from central main column to offset columns
        if upperAttachFlag:
            upperAttachEID = N1.size + 1
            for k in range(ncolumn):
                tempID1 = mainUpperID
                tempID2 = offsetUpperID[k]
                add1, add2 = ghostNodes(nodeMat[tempID1-1,:], nodeMat[tempID2-1,:], rnode[tempID1-1], rnode[tempID2-1])
                if ( (add1[-1] >= z_main[0]) and (add1[-1] <= z_main[-1]) ):
                    tempID1 = xnode.size + 1
                    xnode   = np.append(xnode, add1[0])
                    ynode   = np.append(ynode, add1[1])
                    znode   = np.append(znode, add1[2])
                    gN1     = np.append(gN1, mainUpperID)
                    gN2     = np.append(gN2, tempID1)
                if ( (add2[-1] >= z_offset[0]) and (add2[-1] <= z_offset[-1]) ):
                    tempID2 = xnode.size + 1
                    xnode   = np.append(xnode, add2[0])
                    ynode   = np.append(ynode, add2[1])
                    znode   = np.append(znode, add2[2])
                    gN1     = np.append(gN1, offsetUpperID[k])
                    gN2     = np.append(gN2, tempID2)
                # Pontoon connection
                N1 = np.append(N1, tempID1 )
                N2 = np.append(N2, tempID2 )
                
        # Cross braces from lower central main column to upper offset columns
        if crossAttachFlag:
            crossAttachEID = N1.size + 1
            for k in range(ncolumn):
                tempID1 = mainLowerID
                tempID2 = offsetUpperID[k]
                add1, add2 = ghostNodes(nodeMat[tempID1-1,:], nodeMat[tempID2-1,:], rnode[tempID1-1], rnode[tempID2-1])
                if ( (add1[-1] >= z_main[0]) and (add1[-1] <= z_main[-1]) ):
                    tempID1 = xnode.size + 1
                    xnode   = np.append(xnode, add1[0])
                    ynode   = np.append(ynode, add1[1])
                    znode   = np.append(znode, add1[2])
                    gN1     = np.append(gN1, mainLowerID)
                    gN2     = np.append(gN2, tempID1)
                if ( (add2[-1] >= z_offset[0]) and (add2[-1] <= z_offset[-1]) ):
                    tempID2 = xnode.size + 1
                    xnode   = np.append(xnode, add2[0])
                    ynode   = np.append(ynode, add2[1])
                    znode   = np.append(znode, add2[2])
                    gN1     = np.append(gN1, offsetUpperID[k])
                    gN2     = np.append(gN2, tempID2)
                # Pontoon connection
                N1 = np.append(N1, tempID1 )
                N2 = np.append(N2, tempID2 )
                
            # Will be used later to convert from local member c.s. to global
            cross_angle = np.arctan( (z_attach_upper - z_attach_lower) / R_semi )
            
        # Lower ring around offset columns
        if lowerRingFlag:
            lowerRingEID = N1.size + 1
            for k in range(ncolumn):
                tempID1 = offsetLowerID[k]
                tempID2 = offsetLowerID[k+1]
                add1, add2 = ghostNodes(nodeMat[tempID1-1,:], nodeMat[tempID2-1,:], rnode[tempID1-1], rnode[tempID2-1])
                if ( (add1[-1] >= z_offset[0]) and (add1[-1] <= z_offset[-1]) ):
                    tempID1 = xnode.size + 1
                    xnode   = np.append(xnode, add1[0])
                    ynode   = np.append(ynode, add1[1])
                    znode   = np.append(znode, add1[2])
                    gN1     = np.append(gN1, offsetLowerID[k])
                    gN2     = np.append(gN2, tempID1)
                if ( (add2[-1] >= z_offset[0]) and (add2[-1] <= z_offset[-1]) ):
                    tempID2 = xnode.size + 1
                    xnode   = np.append(xnode, add2[0])
                    ynode   = np.append(ynode, add2[1])
                    znode   = np.append(znode, add2[2])
                    gN1     = np.append(gN1, offsetLowerID[k+1])
                    gN2     = np.append(gN2, tempID2)
                # Pontoon connection
                N1 = np.append(N1, tempID1 )
                N2 = np.append(N2, tempID2 )

        # Upper ring around offset columns
        if upperRingFlag:
            upperRingEID = N1.size + 1
            for k in range(ncolumn):
                tempID1 = offsetUpperID[k]
                tempID2 = offsetUpperID[k+1]
                add1, add2 = ghostNodes(nodeMat[tempID1-1,:], nodeMat[tempID2-1,:], rnode[tempID1-1], rnode[tempID2-1])
                if ( (add1[-1] >= z_offset[0]) and (add1[-1] <= z_offset[-1]) ):
                    tempID1 = xnode.size + 1
                    xnode   = np.append(xnode, add1[0])
                    ynode   = np.append(ynode, add1[1])
                    znode   = np.append(znode, add1[2])
                    gN1     = np.append(gN1, offsetUpperID[k])
                    gN2     = np.append(gN2, tempID1)
                if ( (add2[-1] >= z_offset[0]) and (add2[-1] <= z_offset[-1]) ):
                    tempID2 = xnode.size + 1
                    xnode   = np.append(xnode, add2[0])
                    ynode   = np.append(ynode, add2[1])
                    znode   = np.append(znode, add2[2])
                    gN1     = np.append(gN1, offsetUpperID[k+1])
                    gN2     = np.append(gN2, tempID2)
                # Pontoon connection
                N1 = np.append(N1, tempID1 )
                N2 = np.append(N2, tempID2 )
                
        # Outer cross braces (only one ghost node per connection)
        if outerCrossFlag:
            outerCrossEID = N1.size + 1
            for k in range(ncolumn):
                tempID1 = crossOuterLowerID[k]
                tempID2 = offsetUpperID[k]
                _, add2 = ghostNodes(nodeMat[tempID1-1,:], nodeMat[tempID2-1,:], rnode[tempID1-1], rnode[tempID2-1])
                if ( (add2[-1] >= z_offset[0]) and (add2[-1] <= z_offset[-1]) ):
                    tempID2 = xnode.size + 1
                    xnode   = np.append(xnode, add2[0])
                    ynode   = np.append(ynode, add2[1])
                    znode   = np.append(znode, add2[2])
                    gN1     = np.append(gN1, offsetUpperID[k])
                    gN2     = np.append(gN2, tempID2)
                # Pontoon connection
                N1 = np.append(N1, tempID1 )
                N2 = np.append(N2, tempID2 )

                _, add2 = ghostNodes(nodeMat[crossOuterLowerID[k+1]-1,:], nodeMat[offsetUpperID[k]-1,:], rnode[crossOuterLowerID[k+1]-1], rnode[offsetUpperID[k]-1])
                tempID     = xnode.size + 1
                xnode      = np.append(xnode, add2[0])
                ynode      = np.append(ynode, add2[1])
                znode      = np.append(znode, add2[2])
                gN1        = np.append(gN1, offsetUpperID[k])
                gN2        = np.append(gN2, tempID)
                N1         = np.append(N1, crossOuterLowerID[k+1] )
                N2         = np.append(N2, tempID )

                tempID1 = crossOuterLowerID[k+1]
                tempID2 = offsetUpperID[k]
                _, add2 = ghostNodes(nodeMat[tempID1-1,:], nodeMat[tempID2-1,:], rnode[tempID1-1], rnode[tempID2-1])
                if ( (add2[-1] > z_offset[0]) and (add2[-1] < z_offset[-1]) ):
                    tempID2 = xnode.size + 1
                    xnode   = np.append(xnode, add2[0])
                    ynode   = np.append(ynode, add2[1])
                    znode   = np.append(znode, add2[2])
                    gN1     = np.append(gN1, offsetUpperID[k])
                    gN2     = np.append(gN2, tempID2)
                # Pontoon connection
                N1 = np.append(N1, tempID1 )
                N2 = np.append(N2, tempID2 )
                
            
        # TODO: Parameterize these for upper, lower, cross connections
        # Properties for the inner connectors
        mytube = Tube(2.0*R_od_pontoon, t_wall_pontoon)
        Ax    = mytube.Area * np.ones(N1.shape)
        As    = mytube.Asx  * np.ones(N1.shape)
        Jx    = mytube.J0   * np.ones(N1.shape)
        I     = mytube.Jxx  * np.ones(N1.shape)
        S     = mytube.S    * np.ones(N1.shape)
        C     = mytube.C    * np.ones(N1.shape)
        modE  = E   * np.ones(N1.shape)
        modG  = G   * np.ones(N1.shape)
        roll  = 0.0 * np.ones(N1.shape)
        dens  = rho * np.ones(N1.shape)

        # Add in fairlead support elements
        mooringEID = N1.size + 1
        mytube  = Tube(2.0*R_od_fairlead, t_wall_fairlead)
        for k in range(n_connect):
            kfair   = 0 if ncolumn==0 else k
            
            add1, _ = ghostNodes(nodeMat[fairleadID[kfair]-1,:], nodeMat[mooringID[k]-1,:], rnode[fairleadID[kfair]-1], rnode[mooringID[k]-1])
            tempID  = xnode.size + 1
            xnode   = np.append(xnode, add1[0])
            ynode   = np.append(ynode, add1[1])
            znode   = np.append(znode, add1[2])
            gN1     = np.append(gN1, fairleadID[kfair])
            gN2     = np.append(gN2, tempID)
            N1      = np.append(N1, tempID )
            N2      = np.append(N2, mooringID[k] )

            Ax      = np.append(Ax  , mytube.Area )
            As      = np.append(As  , mytube.Asx )
            Jx      = np.append(Jx  , mytube.J0 )
            I       = np.append(I   , mytube.Jxx )
            S       = np.append(S   , mytube.S )
            C       = np.append(C   , mytube.C )
            modE    = np.append(modE, E )
            modG    = np.append(modG, G )
            roll    = np.append(roll, 0.0 )
            dens    = np.append(dens, rho )

        # Now mock up cylindrical columns as truss members even though long, slender assumption breaks down
        # Will set density = 0.0 so that we don't double count the mass
        # First get geometry in each of the elements
        R_od_main,_     = nodal2sectional( R_od_main )
        R_od_offset,_   = nodal2sectional( R_od_offset )
        R_od_tower,_    = nodal2sectional( R_od_tower )

        # Main column
        mainEID = N1.size + 1
        mytube  = Tube(2.0*R_od_main, t_wall_main)
        myrange = np.arange(R_od_main.size)
        myones  = np.ones(myrange.shape)
        mydens  = m_main / mytube.Area / np.diff(z_main) + eps
        N1   = np.append(N1  , myrange + mainBeginID    )
        N2   = np.append(N2  , myrange + mainBeginID + 1)
        Ax   = np.append(Ax  , mytube.Area )
        As   = np.append(As  , mytube.Asx )
        Jx   = np.append(Jx  , mytube.J0 )
        I    = np.append(I   , mytube.Jxx )
        S    = np.append(S   , mytube.S )
        C    = np.append(C   , mytube.C )
        modE = np.append(modE, E*myones )
        modG = np.append(modG, G*myones )
        roll = np.append(roll, np.zeros(myones.shape) )
        dens = np.append(dens, mydens )

        # Tower column
        towerEID = N1.size + 1
        mytube  = Tube(2.0*R_od_tower, t_wall_tower)
        myrange = np.arange(R_od_tower.size)
        myones  = np.ones(myrange.shape)
        mydens  = m_tower / mytube.Area / np.diff(z_tower) + eps
        N1   = np.append(N1  , myrange + towerBeginID    )
        N2   = np.append(N2  , myrange + towerBeginID + 1)
        Ax   = np.append(Ax  , mytube.Area )
        As   = np.append(As  , mytube.Asx )
        Jx   = np.append(Jx  , mytube.J0 )
        I    = np.append(I   , mytube.Jxx )
        S    = np.append(S   , mytube.S )
        C    = np.append(C   , mytube.C )
        modE = np.append(modE, E*myones )
        modG = np.append(modG, G*myones )
        roll = np.append(roll, np.zeros(myones.shape) )
        dens = np.append(dens, mydens ) 

        # Dummy element
        dummyEID = N1.size + 1
        N1   = np.append(N1  , towerEndID )
        N2   = np.append(N2  , dummyID )
        Ax   = np.append(Ax  , Ax[-1] )
        As   = np.append(As  , As[-1] )
        Jx   = np.append(Jx  , Jx[-1] )
        I    = np.append(I   , I[-1] )
        S    = np.append(S   , S[-1] )
        C    = np.append(C   , C[-1] )
        modE = np.append(modE, 1e20 )
        modG = np.append(modG, 1e20 )
        roll = np.append(roll, 0.0 )
        dens = np.append(dens, 1e-6 ) 

        # Offset column
        offsetEID = []
        mytube     = Tube(2.0*R_od_offset, t_wall_offset)
        myrange    = np.arange(R_od_offset.size)
        myones     = np.ones(myrange.shape)
        mydens     = m_offset / mytube.Area / np.diff(z_offset) + eps
        for k in range(ncolumn):
            offsetEID.append( N1.size + 1 )
            
            N1   = np.append(N1  , myrange + offsetLowerID[k]    )
            N2   = np.append(N2  , myrange + offsetLowerID[k] + 1)
            Ax   = np.append(Ax  , mytube.Area )
            As   = np.append(As  , mytube.Asx )
            Jx   = np.append(Jx  , mytube.J0 )
            I    = np.append(I   , mytube.Jxx )
            S    = np.append(S   , mytube.S )
            C    = np.append(C   , mytube.C )
            modE = np.append(modE, E*myones )
            modG = np.append(modG, G*myones )
            roll = np.append(roll, np.zeros(myones.shape) )
            dens = np.append(dens, mydens ) # Mass added below

        # Ghost elements between centerline nodes and column shells
        ghostEID = N1.size + 1
        myones   = np.ones(gN1.shape)
        N1   = np.append(N1  , gN1 )
        N2   = np.append(N2  , gN2 )
        Ax   = np.append(Ax  , 1e-1*myones )
        As   = np.append(As  , 1e-1*myones )
        Jx   = np.append(Jx  , 1e-1*myones )
        I    = np.append(I   , 1e-1*myones )
        S    = np.append(S   , 1e-1*myones )
        C    = np.append(C   , 1e-1*myones )
        modE = np.append(modE, 1e20*myones )
        modG = np.append(modG, 1e20*myones )
        roll = np.append(roll, 0.0 *myones )
        dens = np.append(dens, 1e-6*myones )

        # Create Node Data object
        nnode   = 1 + np.arange(xnode.size)
        myrnode = np.zeros(xnode.shape) # z-spacing too narrow for use of rnodes
        nodes   = frame3dd.NodeData(nnode, xnode, ynode, znode, myrnode)
        nodeMat = np.c_[xnode, ynode, znode]

        # Create Element Data object
        nelem    = 1 + np.arange(N1.size)
        elements = frame3dd.ElementData(nelem, N1, N2, Ax, As, As, Jx, I, I, modE, modG, roll, dens)

        # Store data for plotting, also handy for operations below
        plotMat = np.zeros((mainEID, 3, 2))
        myn1 = N1[:mainEID]
        myn2 = N2[:mainEID]
        plotMat[:,:,0] = nodeMat[myn1-1,:]
        plotMat[:,:,1] = nodeMat[myn2-1,:]
        discrete_outputs['plot_matrix'] = plotMat
        
        # Compute length and center of gravity for each element for use below
        elemL   = np.sqrt( np.sum( np.diff(plotMat, axis=2)**2.0, axis=1) ).flatten()
        elemCoG = 0.5*np.sum(plotMat, axis=2)
        # Get vertical angle as a measure of welding prep difficulty
        elemAng = np.arccos( np.diff(plotMat[:,-1,:], axis=-1).flatten() / elemL )

        # ---Options object---
        shear = True               # 1: include shear deformation
        geom = False               # 1: include geometric stiffness
        dx = -1                    # x-axis increment for internal forces, -1 to skip
        other = frame3dd.Options(shear, geom, dx)

        # ---LOAD CASES---
        # Extreme loading
        gx = 0.0
        gy = 0.0
        gz = -gravity
        load = frame3dd.StaticLoadCase(gx, gy, gz)

        # Wind + Wave loading in local main / offset / tower c.s.
        Px_main,    Py_main,    Pz_main    = inputs['main_Pz'], inputs['main_Py'], -inputs['main_Px']  # switch to local c.s.
        Px_offset, Py_offset, Pz_offset = inputs['offset_Pz'], inputs['offset_Py'], -inputs['offset_Px']  # switch to local c.s.
        Px_tower,   Py_tower,   Pz_tower   = inputs['tower_Pz'], inputs['tower_Py'], -inputs['tower_Px']  # switch to local c.s.
        epsOff = 1e-5
        # Get mass right- offsets, stiffeners, tower, rna, etc.
        # Also account for buoyancy loads
        # Also apply wind/wave loading as trapezoidal on each element
        # NOTE: Loading is in local element coordinates 0-L, x is along element
        # Base
        nrange  = np.arange(R_od_main.size, dtype=np.int32)
        EL      = mainEID + nrange
        Ux      = F_hydro_main / np.diff(z_main)
        x1 = np.zeros(nrange.shape)
        x2 = np.diff(z_main) - epsOff  # subtract small number b.c. of precision
        wx1, wx2 = Px_main[:-1], Px_main[1:]
        wy1, wy2 = Py_main[:-1], Py_main[1:]
        wz1, wz2 = Pz_main[:-1], Pz_main[1:]
        # Tower
        nrange  = np.arange(R_od_tower.size, dtype=np.int32)
        EL      = np.append(EL, towerEID + nrange)
        Ux      = np.append(Ux, np.zeros(nrange.shape))
        x1      = np.append(x1, np.zeros(nrange.shape))
        x2      = np.append(x2, np.diff(z_tower) - epsOff)
        wx1     = np.append(wx1, Px_tower[:-1])
        wx2     = np.append(wx2, Px_tower[1:])
        wy1     = np.append(wy1, Py_tower[:-1])
        wy2     = np.append(wy2, Py_tower[1:])
        wz1     = np.append(wz1, Pz_tower[:-1])
        wz2     = np.append(wz2, Pz_tower[1:])
        # Buoyancy- offset columns
        nrange  = np.arange(R_od_offset.size, dtype=np.int32)
        for k in range(ncolumn):
            EL      = np.append(EL, offsetEID[k] + nrange)
            Ux      = np.append(Ux,  F_hydro_offset / np.diff(z_offset) )
            x1      = np.append(x1, np.zeros(nrange.shape))
            x2      = np.append(x2, np.diff(z_offset) - epsOff)
            wx1     = np.append(wx1, Px_offset[:-1])
            wx2     = np.append(wx2, Px_offset[1:])
            wy1     = np.append(wy1, Py_offset[:-1])
            wy2     = np.append(wy2, Py_offset[1:])
            wz1     = np.append(wz1, Pz_offset[:-1])
            wz2     = np.append(wz2, Pz_offset[1:])
            
        # Add mass of main and offset columns while we've already done the element enumeration
        Uz = Uy = np.zeros(Ux.shape)
        xx1 = xy1 = xz1 = x1
        xx2 = xy2 = xz2 = x2
        load.changeTrapezoidalLoads(EL, xx1, xx2, wx1, wx2, xy1, xy2, wy1, wy2, xz1, xz2, wz1, wz2)

        # Buoyancy for fully submerged members
        nrange  = np.arange(ncolumn, dtype=np.int32)
        Frange  = np.pi * R_od_pontoon**2 * rhoWater * gravity
        F_truss = 0.0
        z_cb    = np.zeros((3,))
        if ncolumn > 0 and znode[offsetLowerID[0]-1] < 0.0:
            if lowerAttachFlag:
                EL       = np.append(EL, lowerAttachEID + nrange)
                Ux       = np.append(Ux, np.zeros(nrange.shape))
                Uy       = np.append(Uy, np.zeros(nrange.shape))
                Uz       = np.append(Uz, Frange * np.ones(nrange.shape))
                F_truss += Frange * elemL[lowerAttachEID-1] * ncolumn
                z_cb    += Frange * elemL[lowerAttachEID-1] * ncolumn * elemCoG[lowerAttachEID-1,:]
            if lowerRingFlag:
                EL       = np.append(EL, lowerRingEID + nrange)
                Ux       = np.append(Ux, np.zeros(nrange.shape))
                Uy       = np.append(Uy, np.zeros(nrange.shape))
                Uz       = np.append(Uz, Frange * np.ones(nrange.shape))
                F_truss += Frange * elemL[lowerRingEID-1] * ncolumn
                z_cb    += Frange * elemL[lowerRingEID-1] * ncolumn * elemCoG[lowerRingEID-1]
            if crossAttachFlag:
                factor   = np.minimum(1.0, (0.0 - z_attach_lower) / (znode[offsetUpperID[0]-1] - z_attach_lower) )
                EL       = np.append(EL, crossAttachEID + nrange)
                Ux       = np.append(Ux,  factor * Frange * np.sin(cross_angle) * np.ones(nrange.shape))
                Uy       = np.append(Uy, np.zeros(nrange.shape))
                Uz       = np.append(Uz, factor * Frange * np.cos(cross_angle) * np.ones(nrange.shape))
                F_truss += factor * Frange * elemL[crossAttachEID-1] * ncolumn
                z_cb    += factor * Frange * elemL[crossAttachEID-1] * ncolumn * elemCoG[crossAttachEID-1,:]
            if outerCrossFlag:
                factor   = np.minimum(1.0, (0.0 - znode[mainLowerID-1]) / (znode[offsetUpperID[0]-1] - znode[mainLowerID-1]) )
                # TODO: This one will take a little more math
                #EL       = np.append(EL, outerCrossEID + np.arange(2*ncolumn, dtype=np.int32))
                #Ux       = np.append(Ux, np.zeros(nrange.shape))
                #Uy       = np.append(Uy, np.zeros(nrange.shape))
                #Uz       = np.append(Uz, factor * Frange * np.ones(nrange.shape))
                F_truss += factor * Frange * elemL[outerCrossEID-1] * ncolumn
                z_cb    += factor * Frange * elemL[outerCrossEID-1] * ncolumn * elemCoG[outerCrossEID-1,:]
        if ncolumn > 0 and znode[offsetUpperID[0]-1] < 0.0:
            if upperAttachFlag:
                EL       = np.append(EL, upperAttachEID + nrange)
                Ux       = np.append(Ux, np.zeros(nrange.shape))
                Uy       = np.append(Uy, np.zeros(nrange.shape))
                Uz       = np.append(Uz, Frange * np.ones(nrange.shape))
                F_truss += Frange * elemL[upperAttachEID-1] * ncolumn
                z_cb    += Frange * elemL[upperAttachEID-1] * ncolumn * elemCoG[upperAttachEID-1,:]
            if upperRingFlag:
                EL       = np.append(EL, upperRingEID + nrange)
                Ux       = np.append(Ux, np.zeros(nrange.shape))
                Uy       = np.append(Uy, np.zeros(nrange.shape))
                Uz       = np.append(Uz, Frange * np.ones(nrange.shape))
                F_truss += Frange * elemL[upperRingEID-1] * ncolumn
                z_cb    += Frange * elemL[upperRingEID-1] * ncolumn * elemCoG[upperRingEID-1,:]
        # Now do fairlead supports
        nrange   = np.arange(n_connect, dtype=np.int32)
        Frange   = np.pi * R_od_fairlead**2 * rhoWater * gravity
        EL       = np.append(EL, mooringEID + nrange)
        Ux       = np.append(Ux, np.zeros(nrange.shape))
        Uy       = np.append(Uy, np.zeros(nrange.shape))
        Uz       = np.append(Uz, Frange * np.ones(nrange.shape))
        F_truss += Frange * elemL[mooringEID-1] * n_connect
        z_cb    += Frange * elemL[mooringEID-1] * n_connect * elemCoG[mooringEID-1,:]
        # Finally add in all the uniform loads on buoyancy
        load.changeUniformLoads(EL, Ux, Uy, Uz)

        # Point loading for rotor thrust and mooring lines
        # Point loads for mooring loading
        nnode_connect = len(fairleadID)
        nF  = np.array(fairleadID, dtype=np.int32)
        Fx  = np.zeros(nnode_connect)
        Fy  = np.zeros(nnode_connect)
        Fz  = np.zeros(nnode_connect)
        Mxx = np.zeros(nnode_connect)
        Myy = np.zeros(nnode_connect)
        Mzz = np.zeros(nnode_connect)
        for k in range(n_connect):
            iline = 0 if nnode_connect==1 else k
            idx = k*n_lines + np.arange(n_lines)
            Fx[iline] += F_mooring[idx,0].sum()
            Fy[iline] += F_mooring[idx,1].sum()
            Fz[iline] += F_mooring[idx,2].sum()
        # Note: extra momemt from mass accounted for below
        nF  = np.append(nF , towerEndID)
        Fx  = np.append(Fx , F_rna[0] )
        Fy  = np.append(Fy , F_rna[1] )
        Fz  = np.append(Fz , F_rna[2] )
        Mxx = np.append(Mxx, M_rna[0] )
        Myy = np.append(Myy, M_rna[1] )
        Mzz = np.append(Mzz, M_rna[2] )
        
        # Add in all point loads
        load.changePointLoads(nF, Fx, Fy, Fz, Mxx, Myy, Mzz)



        # ---MASS SUMMARIES---
        # Mass summaries now that we've tabulated all of the pontoons
        m_substructure = m_main.sum() + ncolumn*m_offset.sum()
        if mainEID > 1: # Have some pontoons or fairlead supports
            # Buoyancy assembly from incremental calculations above
            V_pontoon = F_truss/rhoWater/gravity
            z_cb      = z_cb[-1] / F_truss if F_truss > 0.0 else 0.0
            outputs['pontoon_displacement'] = V_pontoon
            outputs['pontoon_center_of_buoyancy'] = z_cb

            # Sum up mass and compute CofG.  Frame3DD does mass, but not CG
            ind             = mainEID-1
            m_total         = Ax[:ind] * rho * elemL[:ind]
            m_pontoon       = m_total.sum() #mass.struct_mass
            m_substructure += m_pontoon
            cg_pontoon      = np.sum( m_total[:,np.newaxis] * elemCoG[:ind,:], axis=0 ) / m_total.sum()

            # Compute costs based on "Optimum Design of Steel Structures" by Farkas and Jarmai
            # All dimensions for correlations based on mm, not meters.
            k_m     = inputs['material_cost_rate'] #1.1 # USD / kg carbon steel plate
            k_f     = inputs['labor_cost_rate'] #1.0 # USD / min labor
            k_p     = inputs['painting_cost_rate'] #USD / m^2 painting
            npont   = m_total.size

            # Cost Step 1) Cutting and grinding tube ends
            theta_g = 3.0 # Difficulty factor
            # Cost Step 2) Fillet welds with SMAW (shielded metal arc welding)
            # Multiply by 2 for both ends worth of work
            theta_w = 3.0 # Difficulty factor

            # Labor-based expenses
            K_f = k_f * 2 * ( manufacture.steel_tube_cutgrind_time(theta_g, R_od_pontoon, t_wall_pontoon, elemAng[:ind]) +
                              manufacture.steel_tube_welding_time(theta_w, npont+ncolumn+1, m_substructure, 2*np.pi*R_od_pontoon, t_wall_pontoon) )

            # Cost Step 3) Painting
            theta_p = 2.0
            S_pont  = 2.0 * np.pi * R_od_pontoon * elemL[:ind]
            K_p     = k_p * theta_p * S_pont.sum()

            # Material cost
            K_m = k_m * m_pontoon

            # Total cost
            c_pontoon = K_m + K_f + K_p
            
            outputs['pontoon_mass'] = m_pontoon
            outputs['pontoon_cost'] = c_pontoon
            outputs['pontoon_cost_rate'] = 1e3*c_pontoon/m_pontoon
            outputs['pontoon_center_of_mass'] = cg_pontoon[-1]
        else:
            V_pontoon = z_cb = m_pontoon = 0.0
            cg_pontoon = np.zeros(3)
            
        # Summary of mass and volumes
        outputs['total_displacement'] = V_main.sum() + ncolumn*V_offset.sum() + V_pontoon
        outputs['substructure_mass']  = m_substructure
        outputs['substructure_center_of_mass'] = (ncolumn*m_offset.sum()*cg_offset + m_main.sum()*cg_main +
                                                   m_pontoon*cg_pontoon) / outputs['substructure_mass']
        m_total = outputs['substructure_mass'] + m_rna + m_tower.sum()
        outputs['structural_mass'] = m_total
        outputs['structure_center_of_mass']  = (m_rna*cg_rna + m_tower.sum()*cg_tower +
                                       outputs['substructure_mass']*outputs['substructure_center_of_mass']) / m_total

        # Find cb (center of buoyancy) for whole system
        z_cb = (V_main.sum()*z_cb_main + ncolumn*V_offset.sum()*z_cb_offset + V_pontoon*z_cb) / outputs['total_displacement']
        outputs['z_center_of_buoyancy'] = z_cb


        # ---REACTIONS---
        # Find node closest to CG
        cg_dist = np.sum( (nodeMat - outputs['structure_center_of_mass'][np.newaxis,:])**2, axis=1 )
        cg_node = np.argmin(cg_dist)
        # Free=0, Rigid=inf
        rid = np.array([mainBeginID]) #np.array(fairleadID) #np.array([cg_node+1]) #
        Rx  = np.inf * np.ones(rid.shape)
        Ry  = np.inf * np.ones(rid.shape)
        Rz  = np.inf * np.ones(rid.shape)
        Rxx = np.inf * np.ones(rid.shape)
        Ryy = np.inf * np.ones(rid.shape)
        Rzz = np.inf * np.ones(rid.shape)
        rid = np.append(rid, fairleadID)
        Rx  = np.append(Rx,  K_mooring[0] /nnode_connect * np.ones(nnode_connect) )
        Ry  = np.append(Ry,  K_mooring[1] /nnode_connect * np.ones(nnode_connect) )
        Rz  = np.append(Rz,  K_mooring[2] /nnode_connect * np.ones(nnode_connect) )
        Rxx = np.append(Rxx,  K_mooring[3]/nnode_connect * np.ones(nnode_connect) )
        Ryy = np.append(Ryy,  K_mooring[4]/nnode_connect * np.ones(nnode_connect) )
        Rzz = np.append(Rzz,  K_mooring[5]/nnode_connect * np.ones(nnode_connect) )
        # Get reactions object from frame3dd
        reactions = frame3dd.ReactionData(rid, Rx, Ry, Rz, Rxx, Ryy, Rzz, rigid=np.inf)

        
        # ---FRAME3DD INSTANCE---

        # Initialize frame3dd object
        self.myframe = frame3dd.Frame(nodes, reactions, elements, other)
        
        # Add in extra mass of rna
        inode   = np.array([towerEndID], dtype=np.int32) # rna
        m_extra = np.array([m_rna])
        Ixx = np.array([ I_rna[0] ])
        Iyy = np.array([ I_rna[1] ])
        Izz = np.array([ I_rna[2] ])
        Ixy = np.array([ I_rna[3] ])
        Ixz = np.array([ I_rna[4] ])
        Iyz = np.array([ I_rna[5] ])
        rhox = np.array([ cg_rna[0] ])
        rhoy = np.array([ cg_rna[1] ])
        rhoz = np.array([ cg_rna[2] ])
        self.myframe.changeExtraNodeMass(inode, m_extra, Ixx, Iyy, Izz, Ixy, Ixz, Iyz, rhox, rhoy, rhoz, True)
        
        # Store load case into frame 3dd object
        self.myframe.addLoadCase(load)


        # ---DYNAMIC ANALYSIS---
        # This needs to be compared to FAST until I trust it enough to use it.
        # Have to test BCs, results, mooring stiffness, mooring mass/MOI, etc
        nM = 0 #NFREQ          # number of desired dynamic modes of vibration
        Mmethod = 1         # 1: subspace Jacobi     2: Stodola
        lump = 0            # 0: consistent mass ... 1: lumped mass matrix
        tol = 1e-5          # mode shape tolerance
        shift = 0.0        # shift value ... for unrestrained or partially restrained structures
        
        #self.myframe.enableDynamics(nM, Mmethod, lump, tol, shift)

        # ---DEBUGGING---
        #self.myframe.write('debug.3dd') # For debugging

        # ---RUN ANALYSIS---
        try:
            displacements, forces, reactions, internalForces, mass, modal = self.myframe.run()
        except:
            bad_input()
            return
            
        # --OUTPUTS--
        nE    = nelem.size
        iCase = 0

        # natural frequncies- catch nans and zeros
        temp = np.zeros(NFREQ) #np.array( modal.freq )
        temp[np.isnan(temp)] = 0.0
        outputs['structural_frequencies'] = temp + eps

        # deflections due to loading (from cylinder top and wind/wave loads)
        outputs['top_deflection'] = displacements.dx[iCase, towerEndID-1]  # in yaw-aligned direction

        # Find cg (center of gravity) for whole system
        F_main = -1.0 * np.array([reactions.Fx.sum(), reactions.Fy.sum(), reactions.Fz.sum()])
        M_main = -1.0 * np.array([reactions.Mxx.sum(), reactions.Myy.sum(), reactions.Mzz.sum()])
        r_cg_main = np.array([0.0, 0.0, (znode[mainBeginID] - outputs['structure_center_of_mass'][-1])])
        delta     = np.cross(r_cg_main, F_main)

        outputs['total_force'] = F_main
        outputs['total_moment'] = M_main + delta
        myM= np.cross(np.array([0.0, 0.0, (z_tower[-1] - outputs['structure_center_of_mass'][-1])]), F_rna)
        
        # shear and bending (convert from local to global c.s.)
        Nx = forces.Nx[iCase, 1::2]
        Vy = forces.Vy[iCase, 1::2]
        Vz = forces.Vz[iCase, 1::2]

        Tx = forces.Txx[iCase, 1::2]
        My = forces.Myy[iCase, 1::2]
        Mz = forces.Mzz[iCase, 1::2]

        # Compute axial and shear stresses in elements given Frame3DD outputs and some geomtry data
        # Method comes from Section 7.14 of Frame3DD documentation
        # http://svn.code.sourceforge.net/p/frame3dd/code/trunk/doc/Frame3DD-manual.html#structuralmodeling
        M = np.sqrt(My*My + Mz*Mz)
        sigma_ax = Nx/Ax - M/S
        sigma_sh = np.sqrt(Vy*Vy + Vz*Vz)/As + Tx/C

        # Extract pontoon for stress check
        idx  = range(mainEID-1)
        npon = len(idx)
        if len(idx) > 0:
            qdyn_pontoon = np.max( np.abs( np.r_[inputs['main_qdyn'], inputs['offset_qdyn']] ) )
            sigma_ax_pon = sigma_ax[idx]
            sigma_sh_pon = sigma_sh[idx]
            sigma_h_pon  = util.hoopStress(2*R_od_pontoon, t_wall_pontoon, qdyn_pontoon) * np.ones(sigma_ax_pon.shape)

            outputs['pontoon_stress'][:npon] = util.vonMisesStressUtilization(sigma_ax_pon, sigma_h_pon, sigma_sh_pon,
                                                                               gamma_f*gamma_m*gamma_n, sigma_y)
        
        # Extract tower for Eurocode checks
        idx = towerEID-1 + np.arange(R_od_tower.size, dtype=np.int32)
        L_reinforced   = inputs['tower_buckling_length'] * np.ones(idx.shape)
        sigma_ax_tower = sigma_ax[idx]
        sigma_sh_tower = sigma_sh[idx]
        qdyn_tower,_   = nodal2sectional( inputs['tower_qdyn'] )
        sigma_h_tower  = util.hoopStress(2*R_od_tower, t_wall_tower*np.cos(angle_tower), qdyn_tower)

        outputs['tower_stress:axial'] = sigma_ax_tower
        outputs['tower_stress:shear'] = sigma_sh_tower
        outputs['tower_stress:hoop']  = sigma_h_tower
        outputs['tower_stress:hoopStiffen'] = util.hoopStressEurocode(z_tower, 2*R_od_tower, t_wall_tower, L_reinforced, qdyn_tower)
        outputs['tower_stress'] = util.vonMisesStressUtilization(sigma_ax_tower, sigma_h_tower, sigma_sh_tower,
                                                                  gamma_f*gamma_m*gamma_n, sigma_y)

        sigma_y_vec = sigma_y * np.ones(idx.shape)
        outputs['tower_shell_buckling'] = util.shellBucklingEurocode(2*R_od_tower, t_wall_tower, sigma_ax_tower, sigma_h_tower, sigma_sh_tower,
                                                                      L_reinforced, modE[idx], sigma_y_vec, gamma_f, gamma_b)

        tower_height = z_tower[-1] - z_tower[0]
        outputs['tower_global_buckling'] = util.bucklingGL(2*R_od_tower, t_wall_tower, Nx[idx], M[idx], tower_height, modE[idx], sigma_y_vec, gamma_f, gamma_b)
        
        # Extract main column for Eurocode checks
        idx = mainEID-1 + np.arange(R_od_main.size, dtype=np.int32)
        L_reinforced  = inputs['main_buckling_length']
        sigma_ax_main = sigma_ax[idx]
        sigma_sh_main = sigma_sh[idx]
        qdyn_main,_   = nodal2sectional( inputs['main_qdyn'] )
        sigma_h_main  = util.hoopStress(2*R_od_main, t_wall_main*np.cos(angle_main), qdyn_main)

        outputs['main_stress:axial'] = sigma_ax_main
        outputs['main_stress:shear'] = sigma_sh_main
        outputs['main_stress:hoop']  = sigma_h_main
        outputs['main_stress:hoopStiffen'] = util.hoopStressEurocode(z_main, 2*R_od_main, t_wall_main, L_reinforced, qdyn_main)
        outputs['main_stress'] = util.vonMisesStressUtilization(sigma_ax_main, sigma_h_main, sigma_sh_main,
                                                                        gamma_f*gamma_m*gamma_n, sigma_y)

        sigma_y_vec = sigma_y * np.ones(idx.shape)
        outputs['main_shell_buckling'] = util.shellBucklingEurocode(2*R_od_main, t_wall_main, sigma_ax_main, sigma_h_main, sigma_sh_main,
                                                                            L_reinforced, modE[idx], sigma_y_vec, gamma_f, gamma_b)

        main_height = z_main[-1] - z_main[0]
        outputs['main_global_buckling'] = util.bucklingGL(2*R_od_main, t_wall_main, Nx[idx], M[idx], main_height, modE[idx], sigma_y_vec, gamma_f, gamma_b)

        
        # Extract offset column for Eurocode checks
        if ncolumn > 0:
            idx = offsetEID[0]-1 + np.arange(R_od_offset.size, dtype=np.int32)
            L_reinforced     = inputs['offset_buckling_length']
            sigma_ax_offset = sigma_ax[idx]
            sigma_sh_offset = sigma_sh[idx]
            qdyn_offset,_   = nodal2sectional( inputs['offset_qdyn'] )
            sigma_h_offset  = util.hoopStress(2*R_od_offset, t_wall_offset*np.cos(angle_offset), qdyn_offset)

            outputs['offset_stress:axial'] = sigma_ax_offset
            outputs['offset_stress:shear'] = sigma_sh_offset
            outputs['offset_stress:hoop']  = sigma_h_offset
            outputs['offset_stress:hoopStiffen'] = util.hoopStressEurocode(z_offset, 2*R_od_offset, t_wall_offset, L_reinforced, qdyn_offset)
            outputs['offset_stress'] = util.vonMisesStressUtilization(sigma_ax_offset, sigma_h_offset, sigma_sh_offset,
                                                                            gamma_f*gamma_m*gamma_n, sigma_y)

            sigma_y_vec = sigma_y * np.ones(idx.shape)
            outputs['offset_shell_buckling'] = util.shellBucklingEurocode(2*R_od_offset, t_wall_offset, sigma_ax_offset, sigma_h_offset, sigma_sh_offset,
                                                                                L_reinforced, modE[idx], sigma_y_vec, gamma_f, gamma_b)

            offset_height = z_offset[-1] - z_offset[0]
            outputs['offset_global_buckling'] = util.bucklingGL(2*R_od_offset, t_wall_offset, Nx[idx], M[idx], offset_height, modE[idx], sigma_y_vec, gamma_f, gamma_b)
        
        # TODO: FATIGUE
        # Base and offset columns get API stress/buckling checked in Column Group because that takes into account stiffeners



class TrussIntegerToBoolean(ExplicitComponent):
    def setup(self):
        self.add_input('cross_attachment_pontoons_int', val=1, desc='Inclusion of pontoons that connect the bottom of the central main to the tops of the outer offset columns')
        self.add_input('lower_attachment_pontoons_int', val=1, desc='Inclusion of pontoons that connect the central main to the outer offset columns at their bottoms')
        self.add_input('upper_attachment_pontoons_int', val=1, desc='Inclusion of pontoons that connect the central main to the outer offset columns at their tops')
        self.add_input('lower_ring_pontoons_int', val=1, desc='Inclusion of pontoons that ring around outer offset columns at their bottoms')
        self.add_input('upper_ring_pontoons_int', val=1, desc='Inclusion of pontoons that ring around outer offset columns at their tops')
        self.add_input('outer_cross_pontoons_int', val=1, desc='Inclusion of pontoons that ring around outer offset columns at their tops')

        self.add_discrete_output('cross_attachment_pontoons', val=True, desc='Inclusion of pontoons that connect the bottom of the central main to the tops of the outer offset columns')
        self.add_discrete_output('lower_attachment_pontoons', val=True, desc='Inclusion of pontoons that connect the central main to the outer offset columns at their bottoms')
        self.add_discrete_output('upper_attachment_pontoons', val=True, desc='Inclusion of pontoons that connect the central main to the outer offset columns at their tops')
        self.add_discrete_output('lower_ring_pontoons', val=True, desc='Inclusion of pontoons that ring around outer offset columns at their bottoms')
        self.add_discrete_output('upper_ring_pontoons', val=True, desc='Inclusion of pontoons that ring around outer offset columns at their tops')
        self.add_discrete_output('outer_cross_pontoons', val=True, desc='Inclusion of pontoons that ring around outer offset columns at their tops')

    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
        discrete_outputs['cross_attachment_pontoons'] = (int(inputs['cross_attachment_pontoons_int']) == 1)
        discrete_outputs['lower_attachment_pontoons'] = (int(inputs['lower_attachment_pontoons_int']) == 1)
        discrete_outputs['upper_attachment_pontoons'] = (int(inputs['upper_attachment_pontoons_int']) == 1)
        discrete_outputs['lower_ring_pontoons']       = (int(inputs['lower_ring_pontoons_int']) == 1)
        discrete_outputs['upper_ring_pontoons']       = (int(inputs['upper_ring_pontoons_int']) == 1)
        discrete_outputs['outer_cross_pontoons']      = (int(inputs['outer_cross_pontoons_int']) == 1)

        
# -----------------
#  Assembly
# -----------------

class Loading(Group):

    def initialize(self):
        self.options.declare('nFull')
        self.options.declare('nFullTow')
        
    def setup(self):
        nFull    = self.options['nFull']
        nFullTow = self.options['nFullTow']
        
        # Independent variables that are unique to this Group
        loadingIndeps = IndepVarComp()
        loadingIndeps.add_output('main_pontoon_attach_lower', 0.0)
        loadingIndeps.add_output('main_pontoon_attach_upper', 0.0)
        loadingIndeps.add_output('pontoon_outer_diameter', 0.0, units='m')
        loadingIndeps.add_output('pontoon_wall_thickness', 0.0, units='m')
        loadingIndeps.add_output('outer_cross_pontoons_int', 1)
        loadingIndeps.add_output('cross_attachment_pontoons_int', 1)
        loadingIndeps.add_output('lower_attachment_pontoons_int', 1)
        loadingIndeps.add_output('upper_attachment_pontoons_int', 1)
        loadingIndeps.add_output('lower_ring_pontoons_int', 1)
        loadingIndeps.add_output('upper_ring_pontoons_int', 1)
        loadingIndeps.add_output('connection_ratio_max', 0.0)
        loadingIndeps.add_output('fairlead_support_outer_diameter', 0.0, units='m')
        loadingIndeps.add_output('fairlead_support_wall_thickness', 0.0, units='m')
        self.add_subsystem('loadingIndeps', loadingIndeps, promotes=['*'])
        
        # All the components
        self.add_subsystem('loadingWind', PowerWind(nPoints=nFullTow), promotes=['z0','Uref','shearExp','zref'])
        self.add_subsystem('windLoads', CylinderWindDrag(nPoints=nFullTow), promotes=['cd_usr','beta'])
        self.add_subsystem('intbool', TrussIntegerToBoolean(), promotes=['*'])
        self.add_subsystem('frame', FloatingFrame(nFull=nFull, nFullTow=nFullTow), promotes=['*'])
        
        self.connect('loadingWind.U', 'windLoads.U')
        self.connect('windLoads.windLoads_Px', 'tower_Px')
        self.connect('windLoads.windLoads_Py', 'tower_Py')
        self.connect('windLoads.windLoads_Pz', 'tower_Pz')
        self.connect('windLoads.windLoads_qdyn', 'tower_qdyn')




import os
import numpy as np
import numpy.testing as npt
import unittest
import openmdao.api as om
from wisdem.assemblies.fixed_bottom.monopile_assembly_turbine_nodrive import MonopileTurbine
from wisdem.rotorse.rotor import Init_RotorSE_wRefBlade
from wisdem.rotorse.rotor_geometry_yaml import ReferenceBlade

nSec = 4
nSecTow = 3
NPTS = 100

mydir = os.path.dirname(os.path.realpath(__file__))  # get path to this file

class TestRegression(unittest.TestCase):
    
    def testAssembly(self):
        # Global inputs and outputs
        fname_schema  = mydir + os.sep + 'IEAontology_schema.yaml'
        fname_input   = mydir + os.sep + 'IEA-15-240-RWT.yaml'

        # Initialize blade design
        refBlade = ReferenceBlade()
        refBlade.verbose  = True
        refBlade.NINPUT       = 8
        Nsection_Tow          = 19
        refBlade.NPTS         = 30
        refBlade.spar_var     = ['Spar_cap_ss', 'Spar_cap_ps'] # SS, then PS
        refBlade.te_var       = 'TE_reinforcement'
        refBlade.validate     = False
        refBlade.fname_schema = fname_schema
        blade = refBlade.initialize(fname_input)
        Analysis_Level = 0

        FASTpref                        = {}
        FASTpref['Analysis_Level']      = Analysis_Level
        fst_vt = {}

        prob = om.Problem()
        prob.model=MonopileTurbine(RefBlade=blade, Nsection_Tow=Nsection_Tow, VerbosityCosts=False, FASTpref=FASTpref)
        prob.model.nonlinear_solver = om.NonlinearRunOnce()
        prob.model.linear_solver    = om.DirectSolver()
        prob.setup()

        prob = Init_RotorSE_wRefBlade(prob, blade, Analysis_Level = Analysis_Level, fst_vt = fst_vt)

        # Environmental parameters for the tower
        prob['significant_wave_height'] = 4.52
        prob['significant_wave_period'] = 9.45
        prob['water_depth']             = 30.
        prob['wind_reference_height'] = prob['hub_height'] = 150.
        prob['shearExp']                       = 0.11
        prob['rho']                            = 1.225
        prob['mu']                             = 1.7934e-5
        prob['water_density']                  = 1025.0
        prob['water_viscosity']                = 1.3351e-3
        prob['wind_beta'] = prob['wave_beta'] = 0.0

        # Steel properties for the tower
        prob['material_density']               = 7850.0
        prob['E']                              = 210e9
        prob['G']                              = 79.3e9
        prob['yield_stress']                   = 345e6
        prob['soil_G']                         = 140e6
        prob['soil_nu']                        = 0.4

        # Design constraints
        prob['max_taper_ratio']                = 0.4
        prob['min_diameter_thickness_ratio']   = 120.0

        # Safety factors
        prob['gamma_fatigue']   = 1.755 # (Float): safety factor for fatigue
        prob['gamma_f']         = 1.35  # (Float): safety factor for loads/stresses
        prob['gamma_m']         = 1.3   # (Float): safety factor for materials
        prob['gamma_freq']      = 1.1   # (Float): safety factor for resonant frequencies
        prob['gamma_n']         = 1.0
        prob['gamma_b']         = 1.1

        # Tower
        prob['tower_buckling_length']          = 30.0
        prob['tower_outfitting_factor']        = 1.07
        prob['foundation_height']       = -30.
        prob['suctionpile_depth']       = 45.
        prob['tower_section_height']    = np.array([5., 5., 5., 5., 5., 5., 5., 5., 5., 13.,  13.,  13.,  13.,  13.,  13.,  13.,  13.,  13., 12.58244309])
        prob['tower_outer_diameter'] = np.array([10., 10., 10., 10., 10., 10., 10., 10., 10., 10., 10., 9.92647687, 9.44319282, 8.83283769, 8.15148167, 7.38976138, 6.90908962, 6.74803581, 6.57231775, 6.5])
        prob['tower_wall_thickness'] = np.array([0.05534138, 0.05344902, 0.05150928, 0.04952705, 0.04751736, 0.04551709, 0.0435267, 0.04224176, 0.04105759, 0.0394965, 0.03645589, 0.03377851, 0.03219233, 0.03070819, 0.02910109, 0.02721289, 0.02400931, 0.0208264, 0.02399756])
        prob['tower_buckling_length']   = 15.0
        prob['transition_piece_mass']   = 100e3
        prob['transition_piece_height'] = 15.0

        prob['DC']      = 80.0
        prob['shear']   = True
        prob['geom']    = True
        prob['tower_force_discretization'] = 5.0
        prob['nM']      = 2
        prob['Mmethod'] = 1
        prob['lump']    = 0
        prob['tol']     = 1e-9
        prob['shift']   = 0.0

        # Offshore BOS
        prob['wtiv'] = 'example_wtiv'
        prob['feeder'] = 'future_feeder'
        prob['num_feeders'] = 1
        prob['oss_install_vessel'] = 'example_heavy_lift_vessel'
        prob['site_distance'] = 40.0
        prob['site_distance_to_landfall'] = 40.0
        prob['site_distance_to_interconnection'] = 40.0
        prob['plant_turbine_spacing'] = 7
        prob['plant_row_spacing'] = 7
        prob['plant_substation_distance'] = 1
        prob['tower_deck_space'] = 0.
        prob['nacelle_deck_space'] = 0.
        prob['blade_deck_space'] = 0.
        prob['port_cost_per_month'] = 2e6
        prob['monopile_deck_space'] = 0.
        prob['transition_piece_deck_space'] = 0.
        prob['commissioning_pct'] = 0.01
        prob['decommissioning_pct'] = 0.15
        prob['project_lifetime'] = prob['lifetime'] = 20.0    
        prob['number_of_turbines']             = 40
        prob['annual_opex']                    = 43.56 # $/kW/yr
        #prob['bos_costs']                      = 1234.5 # $/kW

        prob['tower_add_gravity'] = True

        # For turbine costs
        prob['offshore']             = True
        prob['crane']                = False
        prob['bearing_number']       = 2
        prob['crane_cost']           = 0.0
        prob['labor_cost_rate']      = 3.0
        prob['material_cost_rate']   = 2.0
        prob['painting_cost_rate']   = 28.8

        # Drivetrain
        prob['tilt']                    = 6.0
        prob['overhang']                = 11.075
        prob['hub_cm']                  = np.array([-10.685, 0.0, 5.471])
        prob['nac_cm']                  = np.array([-5.718, 0.0, 4.048])
        prob['hub_I']                   = np.array([1382171.187, 2169261.099, 2160636.794, 0.0, 0.0, 0.0])
        prob['nac_I']                   = np.array([13442265.552, 21116729.439, 18382414.385, 0.0, 0.0, 0.0])
        prob['hub_mass']                = 190e3
        prob['nac_mass']                = 797.275e3-190e3
        prob['hss_mass']                = 0.0
        prob['lss_mass']                = 19.504e3
        prob['cover_mass']              = 0.0
        prob['pitch_system_mass']       = 50e3
        prob['platforms_mass']          = 0.0
        prob['spinner_mass']            = 0.0
        prob['transformer_mass']        = 0.0
        prob['vs_electronics_mass']     = 0.0
        prob['yaw_mass']                = 100e3
        prob['gearbox_mass']            = 0.0
        prob['generator_mass']          = 226.7e3+145.25e3
        prob['bedplate_mass']           = 39.434e3
        prob['main_bearing_mass']       = 4.699e3

        prob.run_model()
        # Make sure we get here
        self.assertTrue(True)
        
def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestRegression))
    return suite

if __name__ == '__main__':
    unittest.TextTestRunner().run(suite())
    
    



#-------------------------------------------------------------------------------
# Name:        EmbedLength.py
# Purpose:     This module calculates the necessary embedment length for a pile
#              given soil properties and pile head axial force
#
# Author:      rdamiani
#
# Created:     26/01/2014 - Based entirely on BuildMPtwr.py subroutine under PYTHON/MONOPILE
# Copyright:   (c) rdamiani 2014
# Licence:     <Apache licence>
#-------------------------------------------------------------------------------

import numpy as np
import scipy.integrate, warnings
from constants import gravity

#______________________________________________________________________________#
def EmbedLength(Dpile,tpile,rho,Nhead,soil,gravity=gravity):
      """This function calculates the embedment length for a pile, solely based on \n
        soil friction and toe capacity, no lateral stability. It assumes \n
        both internal and external surfaces available, no plug. \n
        It also does a simple minimization of another function. \n
        INPUT     \n
        fsoil       -float(n), friction coeff. values at z below mudline  \n
        Dpile       -float, pile OD [m]                         \n
        tpile       -float, pile thickness [m]                  \n
        Nhead       -float, axial force [N] at the head of the pile, mudline, ABS value, you may use also the max tensile overall if more conservative, it will be thought as pushing down   \n
        soil        -object of class SoilC
        gravity     -optional, m/s^2 vertical acceleration of gravity (absolute value)
        """

        #From API here is a lookup table
      delta_tab=np.array([15.,20.,25.,30.,35.])  # Soil-pile friction angles
      Nq_tab=np.array([8.,12.,20.,40.,50.])     # Nq values
      f_tab=np.array([47.8,67.,81.3,95.7,114.8])*1.e3     # skin friction limiting values [Pa]
      q_tab=np.array([1.9,2.9,4.8,9.6,12.])*1.e6     # end bearing limiting values [Pa]

      def cu(z,soilzs,cusoil):
        """Function for UNdrained SHear Strength at various levels:\n
        INPUT \n
        z           -float(m), m coordinates below mudline (<0) [m]
        soilzs      -float(n), soil layer bottom coordinates [m]
        cusoil -float(n), Undrained shear strength of layers [N/m2]

        """
        cus=scipy.interp(z,soilzs,cusoil)
        return cus

      def p0(z,soilzs,gammas_soil):
        """Function for UNdrained SHear Strength at various levels:\n
        INPUT \n
        z           -float(m), m coordinates below mudline (<0) [m]
        soilzs      -float(n), soil layer bottom coordinates [m]
        cusoil -float(n), Undrained shear strength of layers [N/m2]


        """
        z=np.asarray(z).reshape([1,-1])  #make sure it is an array

        Dzs=np.hstack((-soilzs[0],(soilzs-np.roll(soilzs,-1))[:-1]))
        weights=gammas_soil*Dzs

        p00=np.zeros(z.size)  #initialize
        for ii in np.arange(0,z.size):
            idx=np.nonzero(soilzs>z[ii])[0] #layers above selected z
            if idx.size:
                p00[ii]=weights[idx].sum()+(soilzs[idx[-1]]-z[ii])*gammas_soil[idx[-1]]

        #Check if within the first layer
        idx=np.nonzero(z>soilzs[0])[0]
        p00[idx]=-z[idx]*gammas_soil[0]

        return p00

      def fsoil(z,delta_soil,SF_pile_ex=soil.SoilSF,sndflg=True):
        """INPUT \n
        z         -float(n),  negative z as in below seabed depth at various n levels [m]
        deltasoil   -float, friction angle between pile and soil [deg] \n

        sndflg  -boolean, True if sand; False if clay
        SF_pile_ex  -float, pile capacity safety factor (API RP2A WSD is 1.5-2, old LFRD 1.25) \n
        """
        z=np.asarray(z).reshape([1,-1])
        if sndflg:
            #K_API   -float, Coefficient of lateral earth pressure (0.8 for unplugged, 1 for plugged piles)
            K_API=0.8 #plug=False
            if soil.plug: #plugged
                K_API=1.
            frsoil=np.min((K_API*np.tan(delta_soil*np.pi/180.)*p0(z,soil.zbots,soil.gammas),scipy.interp(soil.delta,delta_tab,f_tab)))

        else:

            p00=np.zeros(z.size) #initialize
            psi=np.zeros(z.size) #initiali
            alpha=np.zeros(z.size) #initiali

            cu0=cu(z,soil.zbots,soil.cus)

            idx=np.nonzero( z != 0)[0]
            #if not p0(z[idx],soil.zbots,soil.gammas):
            #    print "z=",z[idx]
            #if p0(z[idx],soil.zbots,soil.gammas) ==0.:
            #    print z[idx]
            psi[idx]=cu0/p0(z[idx],soil.zbots,soil.gammas)  #psi for API


            idx2=np.nonzero(psi > 1.)[0]
            idx3=np.nonzero(psi !=0.)[0]
            alpha[idx3]=np.minimum( [1.],[0.5* psi[idx3]**(-.5)])

            alpha[idx2]=np.minimum( [1.],[0.5* psi[idx2]**(-.25)])

            frsoil=alpha*cu0

        return frsoil/SF_pile_ex

        #First integrand is N=int_0^z1 (f*D*pi dz)
      def integrand01(z):
        """INPUT
                fsoil    -float(n), friction coefficient at n z levels
                Dpile    -float or (n), Dmp
                tpile    -float or (n), tmp
                plug     -Boolean, whether or not to consider the pile plugged
        """
        frict=fsoil(z,soil.delta,sndflg=soil.sndflg)
        return frict*np.pi*(Dpile+int(not(soil.plug))*(Dpile-2.*tpile))

      def bearCap(z,Atip,SF_pile_ex=soil.SoilSF,sndflg=True):
        """ Ultimate End-Bearing Capacity of pile. \n
        INPUT
            z         -float(n),  negative z as in below seabed depth at various n levels [m]\n
            Atip  -float, pile tip area [m2]  \n
            SF_pile_ex  -float, pile capacity safety factor  \n
            sndflg  -boolean, True if sand; False if clay
            """
        if sndflg:
            bearcap=Atip*np.min((p0(z,soil.zbots,soil.gammas)*scipy.interp(soil.delta,delta_tab,Nq_tab),scipy.interp(soil.delta,delta_tab,q_tab)))
        else:
            bearcap=9.*Atip*cu(z,soil.zbots,soil.cus)
        return bearcap/SF_pile_ex

      # I need to solve an integral equation
      #let us define the function that needs to be minimized to 0
      def minfun(zembd):  #function to minimize
        Atip=np.pi/4. * (Dpile**2. -(Dpile-2.*tpile)**2.)  #x-sect area [m2]
        wght_embd=-Atip*zembd * rho*gravity # weight of the embedded pile
        Qp=bearCap(zembd,Atip,sndflg=soil.sndflg)

        fun2min=scipy.integrate.romberg(integrand01,zembd,0.,divmax=20)    \
        -wght_embd-Nhead #+Qp
        return fun2min[0] #Note not using Qp to simulate pullout also

      #Solve teh equation here

      #print "p0=",p0(-0.00091553,soil.zbots,soil.gammas)
      #print "minfun(-30.)=",minfun(-30.),"fsoil=",fsoil(-30.,soil.delta,sndflg=soil.sndflg) #debug

     # Qp=bearCap(-30.,1.02,sndflg=soil.sndflg); debug
      guess=-15.  #embedment length
      res=scipy.optimize.fsolve(minfun,guess,full_output=1)
      if res[-2] != 1:
            warnings.warn('Embedment Length not found')
      return res[0][0] #zembd
#______________________________________________________________________________#

if __name__ == '__main__':
    main()




import numpy as np
from openmdao.api import ExplicitComponent
from wisdem.commonse.tube import CylindricalShellProperties

from wisdem.commonse import gravity, eps
import wisdem.commonse.frustum as frustum
import wisdem.commonse.manufacturing as manufacture
from wisdem.commonse.UtilizationSupplement import hoopStressEurocode, hoopStress
from wisdem.commonse.utilities import assembleI, unassembleI, sectionalInterp, nodal2sectional
import wisdem.pyframe3dd.frame3dd as frame3dd

RIGID = 1e30

# -----------------
#  Components
# -----------------

#TODO need to check the length of each array
class CylinderDiscretization(ExplicitComponent):
    """discretize geometry into finite element nodes"""

    def initialize(self):
        self.options.declare('nPoints')
        self.options.declare('nRefine')

    def setup(self):
        nPoints = self.options['nPoints']
        nRefine = np.round( self.options['nRefine'] )
        nFull   = int(nRefine * (nPoints-1) + 1)
        
        # variables
        self.add_input('foundation_height', val=0.0, units='m', desc='starting height of tower') 
        self.add_input('section_height', np.zeros(nPoints-1), units='m', desc='parameterized section heights along cylinder')
        self.add_input('diameter', np.zeros(nPoints), units='m', desc='cylinder diameter at corresponding locations')
        self.add_input('wall_thickness', np.zeros(nPoints-1), units='m', desc='shell thickness at corresponding locations')

        #out
        self.add_output('z_param', np.zeros(nPoints), units='m', desc='parameterized locations along cylinder, linear lofting between')
        self.add_output('z_full', np.zeros(nFull), units='m', desc='locations along cylinder')
        self.add_output('d_full', np.zeros(nFull), units='m', desc='cylinder diameter at corresponding locations')
        self.add_output('t_full', np.zeros(nFull-1), units='m', desc='shell thickness at corresponding locations')
        # Convenience outputs for export to other modules
        
        # Derivatives
        # self.declare_partials('*', '*', method='fd', form='central', step=1e-6)

    def compute(self, inputs, outputs):
        nRefine = int(np.round( self.options['nRefine'] ))
        z_param = float(inputs['foundation_height']) + np.r_[0.0, np.cumsum(inputs['section_height'].flatten())]
        # Have to regine each element one at a time so that we preserve input nodes
        z_full = np.array([])
        for k in range(z_param.size-1):
            zref = np.linspace(z_param[k], z_param[k+1], nRefine+1)
            z_full = np.append(z_full, zref)
        z_full = np.unique(z_full)
        outputs['z_full']  = z_full
        outputs['d_full']  = np.interp(z_full, z_param, inputs['diameter'])
        z_section = 0.5*(z_full[:-1] + z_full[1:])
        outputs['t_full']  = sectionalInterp(z_section, z_param, inputs['wall_thickness'])
        outputs['z_param'] = z_param

class CylinderMass(ExplicitComponent):

    def initialize(self):
        self.options.declare('nPoints')
        
    def setup(self):
        nPoints = self.options['nPoints']
        
        self.add_input('d_full', val=np.zeros(nPoints), units='m', desc='cylinder diameter at corresponding locations')
        self.add_input('t_full', val=np.zeros(nPoints-1), units='m', desc='shell thickness at corresponding locations')
        self.add_input('z_full', val=np.zeros(nPoints), units='m', desc='parameterized locations along cylinder, linear lofting between')
        self.add_input('material_density', 0.0, units='kg/m**3', desc='material density')
        self.add_input('outfitting_factor', val=0.0, desc='Multiplier that accounts for secondary structure mass inside of cylinder')
        
        self.add_input('material_cost_rate', 0.0, units='USD/kg', desc='Raw material cost rate: steel $1.1/kg, aluminum $3.5/kg')
        self.add_input('labor_cost_rate', 0.0, units='USD/min', desc='Labor cost rate')
        self.add_input('painting_cost_rate', 0.0, units='USD/m/m', desc='Painting / surface finishing cost rate')
        
        self.add_output('cost', val=0.0, units='USD', desc='Total cylinder cost')
        self.add_output('mass', val=np.zeros(nPoints-1), units='kg', desc='Total cylinder mass')
        self.add_output('center_of_mass', val=0.0, units='m', desc='z-position of center of mass of cylinder')
        self.add_output('section_center_of_mass', val=np.zeros(nPoints-1), units='m', desc='z position of center of mass of each can in the cylinder')
        self.add_output('I_base', np.zeros((6,)), units='kg*m**2', desc='mass moment of inertia of cylinder about base [xx yy zz xy xz yz]')
        
        # Derivatives
        # self.declare_partials('*', '*', method='fd', form='central', step=1e-6)
        
    def compute(self, inputs, outputs):
        # Unpack variables for thickness and average radius at each can interface
        twall = inputs['t_full']
        Rb    = 0.5*inputs['d_full'][:-1]
        Rt    = 0.5*inputs['d_full'][1:]
        zz    = inputs['z_full']
        H     = np.diff(zz)
        rho   = inputs['material_density']
        coeff = inputs['outfitting_factor']
        if coeff < 1.0: coeff += 1.0

        # Total mass of cylinder
        V_shell = frustum.frustumShellVol(Rb, Rt, twall, H)
        outputs['mass'] = coeff * rho * V_shell
        
        # Center of mass of each can/section
        cm_section = zz[:-1] + frustum.frustumShellCG(Rb, Rt, twall, H)
        outputs['section_center_of_mass'] = cm_section

        # Center of mass of cylinder
        V_shell += eps
        outputs['center_of_mass'] = np.dot(V_shell, cm_section) / V_shell.sum()

        # Moments of inertia
        Izz_section = frustum.frustumShellIzz(Rb, Rt, twall, H)
        Ixx_section = Iyy_section = frustum.frustumShellIxx(Rb, Rt, twall, H)

        # Sum up each cylinder section using parallel axis theorem
        I_base = np.zeros((3,3))
        for k in range(Izz_section.size):
            R = np.array([0.0, 0.0, cm_section[k]])
            Icg = assembleI( [Ixx_section[k], Iyy_section[k], Izz_section[k], 0.0, 0.0, 0.0] )

            I_base += Icg + V_shell[k]*(np.dot(R, R)*np.eye(3) - np.outer(R, R))
            
        # All of the mass and volume terms need to be multiplied by density
        I_base *= coeff * rho

        outputs['I_base'] = unassembleI(I_base)
        

        # Compute costs based on "Optimum Design of Steel Structures" by Farkas and Jarmai
        # All dimensions for correlations based on mm, not meters.
        R_ave  = 0.5*(Rb + Rt)
        taper  = np.minimum(Rb/Rt, Rt/Rb)
        nsec   = twall.size
        mplate = rho * V_shell.sum()
        k_m    = inputs['material_cost_rate'] #1.1 # USD / kg carbon steel plate
        k_f    = inputs['labor_cost_rate'] #1.0 # USD / min labor
        k_p    = inputs['painting_cost_rate'] #USD / m^2 painting
        
        # Cost Step 1) Cutting flat plates for taper using plasma cutter
        cutLengths = 2.0 * np.sqrt( (Rt-Rb)**2.0 + H**2.0 ) # Factor of 2 for both sides
        # Cost Step 2) Rolling plates 
        # Cost Step 3) Welding rolled plates into shells (set difficulty factor based on tapering with logistic function)
        theta_F = 4.0 - 3.0 / (1 + np.exp(-5.0*(taper-0.75)))
        # Cost Step 4) Circumferential welds to join cans together
        theta_A = 2.0

        # Labor-based expenses
        K_f = k_f * ( manufacture.steel_cutting_plasma_time(cutLengths, twall) +
                      manufacture.steel_rolling_time(theta_F, R_ave, twall) +
                      manufacture.steel_butt_welding_time(theta_A, nsec, mplate, cutLengths, twall) +
                      manufacture.steel_butt_welding_time(theta_A, nsec, mplate, 2*np.pi*Rb[1:], twall[1:]) )
        
        # Cost step 5) Painting- outside and inside
        theta_p = 2
        K_p  = k_p * theta_p * 2 * (2 * np.pi * R_ave * H).sum()

        # Cost step 6) Outfitting
        K_o = 1.5 * k_m * (coeff - 1.0) * mplate
        
        # Material cost, without outfitting
        K_m = k_m * mplate

        # Assemble all costs
        outputs['cost'] = K_m + K_o + K_p + K_f

        

#@implement_base(CylinderFromCSProps)
class CylinderFrame3DD(ExplicitComponent):
    def initialize(self):
        self.options.declare('npts')
        self.options.declare('nK')
        self.options.declare('nMass')
        self.options.declare('nPL')
        
    def setup(self):
        npts  = self.options['npts']
        nK    = self.options['nK']
        nMass = self.options['nMass']
        nPL   = self.options['nPL']

        # cross-sectional data along cylinder.
        self.add_input('z', np.zeros(npts), units='m', desc='location along cylinder. start at bottom and go to top')
        self.add_input('Az', np.zeros(npts-1), units='m**2', desc='cross-sectional area')
        self.add_input('Asx', np.zeros(npts-1), units='m**2', desc='x shear area')
        self.add_input('Asy', np.zeros(npts-1), units='m**2', desc='y shear area')
        self.add_input('Jz', np.zeros(npts-1), units='m**4', desc='polar moment of inertia')
        self.add_input('Ixx', np.zeros(npts-1), units='m**4', desc='area moment of inertia about x-axis')
        self.add_input('Iyy', np.zeros(npts-1), units='m**4', desc='area moment of inertia about y-axis')

        self.add_input('E', val=0.0, units='N/m**2', desc='modulus of elasticity')
        self.add_input('G', val=0.0, units='N/m**2', desc='shear modulus')
        self.add_input('rho', val=0.0, units='kg/m**3', desc='material density')
        self.add_input('sigma_y', val=0.0, units='N/m**2', desc='yield stress')
        self.add_input('L_reinforced', val=0.0, units='m')

        # effective geometry -- used for handbook methods to estimate hoop stress, buckling, fatigue
        self.add_input('d', np.zeros(npts), units='m', desc='effective cylinder diameter for section')
        self.add_input('t', np.zeros(npts-1), units='m', desc='effective shell thickness for section')

        # spring reaction data.  Use global RIGID for rigid constraints.
        self.add_input('kidx', np.zeros(nK, dtype=np.int_), desc='indices of z where external stiffness reactions should be applied.')
        self.add_input('kx', np.zeros(nK), units='N/m', desc='spring stiffness in x-direction')
        self.add_input('ky', np.zeros(nK), units='N/m', desc='spring stiffness in y-direction')
        self.add_input('kz', np.zeros(nK), units='N/m', desc='spring stiffness in z-direction')
        self.add_input('ktx', np.zeros(nK), units='N/m', desc='spring stiffness in theta_x-rotation')
        self.add_input('kty', np.zeros(nK), units='N/m', desc='spring stiffness in theta_y-rotation')
        self.add_input('ktz', np.zeros(nK), units='N/m', desc='spring stiffness in theta_z-rotation')

        # extra mass
        self.add_input('midx', np.zeros(nMass, dtype=np.int_), desc='indices where added mass should be applied.')
        self.add_input('m', np.zeros(nMass), units='kg', desc='added mass')
        self.add_input('mIxx', np.zeros(nMass), units='kg*m**2', desc='x mass moment of inertia about some point p')
        self.add_input('mIyy', np.zeros(nMass), units='kg*m**2', desc='y mass moment of inertia about some point p')
        self.add_input('mIzz', np.zeros(nMass), units='kg*m**2', desc='z mass moment of inertia about some point p')
        self.add_input('mIxy', np.zeros(nMass), units='kg*m**2', desc='xy mass moment of inertia about some point p')
        self.add_input('mIxz', np.zeros(nMass), units='kg*m**2', desc='xz mass moment of inertia about some point p')
        self.add_input('mIyz', np.zeros(nMass), units='kg*m**2', desc='yz mass moment of inertia about some point p')
        self.add_input('mrhox', np.zeros(nMass), units='m', desc='x-location of p relative to node')
        self.add_input('mrhoy', np.zeros(nMass), units='m', desc='y-location of p relative to node')
        self.add_input('mrhoz', np.zeros(nMass), units='m', desc='z-location of p relative to node')
        self.add_discrete_input('addGravityLoadForExtraMass', True, desc='add gravitational load')

        # point loads (if addGravityLoadForExtraMass=True be sure not to double count by adding those force here also)
        self.add_input('plidx', np.zeros(nPL, dtype=np.int_), desc='indices where point loads should be applied.')
        self.add_input('Fx', np.zeros(nPL), units='N', desc='point force in x-direction')
        self.add_input('Fy', np.zeros(nPL), units='N', desc='point force in y-direction')
        self.add_input('Fz', np.zeros(nPL), units='N', desc='point force in z-direction')
        self.add_input('Mxx', np.zeros(nPL), units='N*m', desc='point moment about x-axis')
        self.add_input('Myy', np.zeros(nPL), units='N*m', desc='point moment about y-axis')
        self.add_input('Mzz', np.zeros(nPL), units='N*m', desc='point moment about z-axis')

        # combined wind-water distributed loads
        self.add_input('Px', np.zeros(npts), units='N/m', desc='force per unit length in x-direction')
        self.add_input('Py', np.zeros(npts), units='N/m', desc='force per unit length in y-direction')
        self.add_input('Pz', np.zeros(npts), units='N/m', desc='force per unit length in z-direction')
        self.add_input('qdyn', np.zeros(npts), units='N/m**2', desc='dynamic pressure')

        # options
        self.add_discrete_input('shear', True, desc='include shear deformation')
        self.add_discrete_input('geom', False, desc='include geometric stiffness')
        self.add_input('dx', 5.0, desc='z-axis increment for internal forces')
        self.add_discrete_input('nM', 2, desc='number of desired dynamic modes of vibration (below only necessary if nM > 0)')
        self.add_discrete_input('Mmethod', 1, desc='1: subspace Jacobi, 2: Stodola')
        self.add_discrete_input('lump', 0, desc='0: consistent mass, 1: lumped mass matrix')
        self.add_input('tol', 1e-9, desc='mode shape tolerance')
        self.add_input('shift', 0.0, desc='shift value ... for unrestrained structures')

        # outputs
        self.add_output('mass', 0.0)
        self.add_output('f1', 0.0, units='Hz', desc='First natural frequency')
        self.add_output('f2', 0.0, units='Hz', desc='Second natural frequency')
        self.add_output('top_deflection', 0.0, units='m', desc='Deflection of cylinder top in yaw-aligned +x direction')
        self.add_output('Fz_out', np.zeros(npts-1), units='N', desc='Axial foce in vertical z-direction in cylinder structure.')
        self.add_output('Vx_out', np.zeros(npts-1), units='N', desc='Shear force in x-direction in cylinder structure.')
        self.add_output('Vy_out', np.zeros(npts-1), units='N', desc='Shear force in y-direction in cylinder structure.')
        self.add_output('Mxx_out', np.zeros(npts-1), units='N*m', desc='Moment about x-axis in cylinder structure.')
        self.add_output('Myy_out', np.zeros(npts-1), units='N*m', desc='Moment about y-axis in cylinder structure.')
        self.add_output('Mzz_out', np.zeros(npts-1), units='N*m', desc='Moment about z-axis in cylinder structure.')
        self.add_output('base_F', val=np.zeros(3), units='N', desc='Total force on cylinder')
        self.add_output('base_M', val=np.zeros(3), units='N*m', desc='Total moment on cylinder measured at base')

        self.add_output('axial_stress', np.zeros(npts-1), units='N/m**2', desc='Axial stress in cylinder structure')
        self.add_output('shear_stress', np.zeros(npts-1), units='N/m**2', desc='Shear stress in cylinder structure')
        self.add_output('hoop_stress', np.zeros(npts-1), units='N/m**2', desc='Hoop stress in cylinder structure calculated with simple method used in API standards')
        self.add_output('hoop_stress_euro', np.zeros(npts-1), units='N/m**2', desc='Hoop stress in cylinder structure calculated with Eurocode method')
        
        # Derivatives
        # self.declare_partials('*', '*', method='fd', form='central', step=1e-6)

        
    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):

        # ------- node data ----------------
        z = inputs['z']
        n = len(z)
        node = np.arange(1, n+1)
        x = np.zeros(n)
        y = np.zeros(n)
        r = np.zeros(n)

        nodes = frame3dd.NodeData(node, x, y, z, r)
        # -----------------------------------

        # ------ reaction data ------------

        # rigid base
        node = inputs['kidx'] + np.ones(len(inputs['kidx']))   # add one because 0-based index but 1-based node numbering
        rigid = RIGID

        reactions = frame3dd.ReactionData(node, inputs['kx'], inputs['ky'], inputs['kz'], inputs['ktx'], inputs['kty'], inputs['ktz'], rigid)
        # -----------------------------------

        # ------ frame element data ------------
        element = np.arange(1, n)
        N1 = np.arange(1, n)
        N2 = np.arange(2, n+1)

        roll = np.zeros(n-1)

        # average across element b.c. frame3dd uses constant section elements
        # TODO: Use nodal2sectional
        Az  = inputs['Az']
        Asx = inputs['Asx']
        Asy = inputs['Asy']
        Jz  = inputs['Jz']
        Ixx = inputs['Ixx']
        Iyy = inputs['Iyy']
        E   = inputs['E']*np.ones(Az.shape)
        G   = inputs['G']*np.ones(Az.shape)
        rho = inputs['rho']*np.ones(Az.shape)

        elements = frame3dd.ElementData(element, N1, N2, Az, Asx, Asy, Jz, Ixx, Iyy, E, G, roll, rho)
        # -----------------------------------


        # ------ options ------------
        options = frame3dd.Options(discrete_inputs['shear'], discrete_inputs['geom'], float(inputs['dx']))
        # -----------------------------------

        # initialize frame3dd object
        cylinder = frame3dd.Frame(nodes, reactions, elements, options)


        # ------ add extra mass ------------

        # extra node inertia data
        N = inputs['midx'] + np.ones(len(inputs['midx']))

        cylinder.changeExtraNodeMass(N, inputs['m'], inputs['mIxx'], inputs['mIyy'], inputs['mIzz'], inputs['mIxy'], inputs['mIxz'], inputs['mIyz'],
            inputs['mrhox'], inputs['mrhoy'], inputs['mrhoz'], discrete_inputs['addGravityLoadForExtraMass'])

        # ------------------------------------

        # ------- enable dynamic analysis ----------
        cylinder.enableDynamics(discrete_inputs['nM'], discrete_inputs['Mmethod'], discrete_inputs['lump'], float(inputs['tol']), float(inputs['shift']))
        # ----------------------------

        # ------ static load case 1 ------------

        # gravity in the X, Y, Z, directions (global)
        gx = 0.0
        gy = 0.0
        gz = -gravity

        load = frame3dd.StaticLoadCase(gx, gy, gz)

        # point loads
        nF = inputs['plidx'] + np.ones(len(inputs['plidx']))
        load.changePointLoads(nF, inputs['Fx'], inputs['Fy'], inputs['Fz'], inputs['Mxx'], inputs['Myy'], inputs['Mzz'])

        # distributed loads
        Px, Py, Pz = inputs['Pz'], inputs['Py'], -inputs['Px']  # switch to local c.s.
        z = inputs['z']

        # trapezoidally distributed loads
        EL = np.arange(1, n)
        xx1 = xy1 = xz1 = np.zeros(n-1)
        xx2 = xy2 = xz2 = np.diff(z) - 1e-6  # subtract small number b.c. of precision
        wx1 = Px[:-1]
        wx2 = Px[1:]
        wy1 = Py[:-1]
        wy2 = Py[1:]
        wz1 = Pz[:-1]
        wz2 = Pz[1:]

        load.changeTrapezoidalLoads(EL, xx1, xx2, wx1, wx2, xy1, xy2, wy1, wy2, xz1, xz2, wz1, wz2)

        cylinder.addLoadCase(load)
        # Debugging
        #cylinder.write('temp.3dd')
        # -----------------------------------
        # run the analysis
        displacements, forces, reactions, internalForces, mass, modal = cylinder.run()
        iCase = 0

        # mass
        outputs['mass'] = mass.struct_mass

        # natural frequncies
        outputs['f1'] = modal.freq[0]
        outputs['f2'] = modal.freq[1]

        # deflections due to loading (from cylinder top and wind/wave loads)
        outputs['top_deflection'] = displacements.dx[iCase, n-1]  # in yaw-aligned direction

        # shear and bending, one per element (convert from local to global c.s.)
        Fz = forces.Nx[iCase, 1::2]
        Vy = forces.Vy[iCase, 1::2]
        Vx = -forces.Vz[iCase, 1::2]

        Mzz = forces.Txx[iCase, 1::2]
        Myy = forces.Myy[iCase, 1::2]
        Mxx = -forces.Mzz[iCase, 1::2]

        # Record total forces and moments
        outputs['base_F'] = -1.0 * np.array([reactions.Fx.sum(), reactions.Fy.sum(), reactions.Fz.sum()])
        outputs['base_M'] = -1.0 * np.array([reactions.Mxx.sum(), reactions.Myy.sum(), reactions.Mzz.sum()])

        outputs['Fz_out']  = Fz
        outputs['Vx_out']  = Vx
        outputs['Vy_out']  = Vy
        outputs['Mxx_out'] = Mxx
        outputs['Myy_out'] = Myy
        outputs['Mzz_out'] = Mzz

        # axial and shear stress
        d,_    = nodal2sectional(inputs['d'])
        qdyn,_ = nodal2sectional(inputs['qdyn'])
        
        ##R = self.d/2.0
        ##x_stress = R*np.cos(self.theta_stress)
        ##y_stress = R*np.sin(self.theta_stress)
        ##axial_stress = Fz/self.Az + Mxx/self.Ixx*y_stress - Myy/self.Iyy*x_stress
#        V = Vy*x_stress/R - Vx*y_stress/R  # shear stress orthogonal to direction x,y
#        shear_stress = 2. * V / self.Az  # coefficient of 2 for a hollow circular section, but should be conservative for other shapes
        outputs['axial_stress'] = Fz/inputs['Az'] - np.sqrt(Mxx**2+Myy**2)/inputs['Iyy']*d/2.0  #More conservative, just use the tilted bending and add total max shear as well at the same point, if you do not like it go back to the previous lines

        outputs['shear_stress'] = 2. * np.sqrt(Vx**2+Vy**2) / inputs['Az'] # coefficient of 2 for a hollow circular section, but should be conservative for other shapes

        # hoop_stress (Eurocode method)
        L_reinforced = inputs['L_reinforced'] * np.ones(Fz.shape)
        outputs['hoop_stress_euro'] = hoopStressEurocode(inputs['z'], d, inputs['t'], L_reinforced, qdyn)

        # Simpler hoop stress used in API calculations
        outputs['hoop_stress'] = hoopStress(d, inputs['t'], qdyn)



#!/usr/bin/env python
# encoding: utf-8
"""
utilities.py

Created by Andrew Ning on 2013-05-31.
Copyright (c) NREL. All rights reserved.
"""

from __future__ import print_function
import numpy as np
from scipy.linalg import solve_banded


def cosd(value):
    """cosine of value where value is given in degrees"""

    return np.cos(np.radians(value))


def sind(value):
    """sine of value where value is given in degrees"""

    return np.sin(np.radians(value))


def tand(value):
    """tangent of value where value is given in degrees"""

    return np.tan(np.radians(value))


def hstack(vec):
    """stack arrays horizontally.  useful for assemblying Jacobian
    assumes arrays are column vectors (if rows just use concatenate)"""
    newvec = []
    for v in vec:
        if len(v.shape) == 1:
            newvec.append(v[:, np.newaxis])
        else:
            newvec.append(v)

    return np.hstack(newvec)


def vstack(vec):
    """stack arrays vertically
    assumes arrays are row vectors.  if columns use concatenate"""

    newvec = []
    for v in vec:
        if len(v.shape) == 1:
            newvec.append(v[np.newaxis, :])
        else:
            newvec.append(v)

    return np.vstack(newvec)


def _checkIfFloat(x):
    try:
        n = len(x)
    except TypeError:  # if x is just a float
        x = np.array([x])
        n = 1

    return x, n


def linspace_with_deriv(start, stop, num):
    """creates linearly spaced arrays, and derivatives for changing end points"""

    step = (stop-start)/float((num-1))
    y = np.arange(0, num) * step + start
    y[-1] = stop

    # gradients
    const = np.arange(0, num) * 1.0/float((num-1))
    dy_dstart = -const + 1.0
    dy_dstart[-1] = 0.0

    dy_dstop = const
    dy_dstop[-1] = 1.0

    return y, dy_dstart, dy_dstop



def sectionalInterp(xi, x, y):
    epsilon = 1e-11
    xx=np.c_[x[:-1], x[1:]-epsilon].flatten()
    yy=np.c_[y, y].flatten()    
    return np.interp(xi, xx, yy)


def interp_with_deriv(x, xp, yp):
    """linear interpolation and its derivative. To be precise, linear interpolation is not
    differentiable right at the control points, but in general it works well enough"""
    # TODO: put in Fortran to speed up

    x, n = _checkIfFloat(x)

    if np.any(np.diff(xp) < 0):
        raise TypeError('xp must be in ascending order')

    # n = len(x)
    m = len(xp)

    y = np.zeros(n)
    dydx = np.zeros(n)
    dydxp = np.zeros((n, m))
    dydyp = np.zeros((n, m))

    for i in range(n):
        if x[i] < xp[0]:
            j = 0  # linearly extrapolate
        elif x[i] > xp[-1]:
            j = m-2
        else:
            for j in range(m-1):
                if xp[j+1] > x[i]:
                    break
        x1 = xp[j]
        y1 = yp[j]
        x2 = xp[j+1]
        y2 = yp[j+1]

        y[i] = y1 + (y2 - y1)*(x[i] - x1)/(x2 - x1)
        dydx[i] = (y2 - y1)/(x2 - x1)
        dydxp[i, j] = (y2 - y1)*(x[i] - x2)/(x2 - x1)**2
        dydxp[i, j+1] = -(y2 - y1)*(x[i] - x1)/(x2 - x1)**2
        dydyp[i, j] = 1 - (x[i] - x1)/(x2 - x1)
        dydyp[i, j+1] = (x[i] - x1)/(x2 - x1)

    if n == 1:
        y = y[0]

    return y, np.diag(dydx), dydxp, dydyp


def assembleI(I):
    Ixx, Iyy, Izz, Ixy, Ixz, Iyz = I[0], I[1], I[2], I[3], I[4], I[5] 
    return np.array([[Ixx, Ixy, Ixz], [Ixy, Iyy, Iyz], [Ixz, Iyz, Izz]])

def unassembleI(I):
    return np.array([I[0, 0], I[1, 1], I[2, 2], I[0, 1], I[0, 2], I[1, 2]])


def cubic_with_deriv(x, xp, yp):
    """deprecated"""

    x, n = _checkIfFloat(x)

    if np.any(np.diff(xp) < 0):
        raise TypeError('xp must be in ascending order')

    # n = len(x)
    m = len(xp)

    y = np.zeros(n)
    dydx = np.zeros(n)
    dydxp = np.zeros((n, m))
    dydyp = np.zeros((n, m))

    xk = xp[1:-1]
    yk = yp[1:-1]
    xkp = xp[2:]
    ykp = yp[2:]
    xkm = xp[:-2]
    ykm = yp[:-2]

    b = (ykp - yk)/(xkp - xk) - (yk - ykm)/(xk - xkm)
    l = (xk - xkm)/6.0
    d = (xkp - xkm)/3.0
    u = (xkp - xk)/6.0
    # u[0] = 0.0  # non-existent entries
    # l[-1] = 0.0

    # solve for second derivatives
    fpp = solve_banded((1, 1), np.array([u, d, l]), b)
    fpp = np.concatenate([[0.0], fpp, [0.0]])  # natural spline

    # find location in vector
    for i in range(n):
        if x[i] < xp[0]:
            j = 0
        elif x[i] > xp[-1]:
            j = m-2
        else:
            for j in range(m-1):
                if xp[j+1] > x[i]:
                    break
        x1 = xp[j]
        y1 = yp[j]
        x2 = xp[j+1]
        y2 = yp[j+1]

        A = (x2 - x[i])/(x2 - x1)
        B = 1 - A
        C = 1.0/6*(A**3 - A)*(x2 - x1)**2
        D = 1.0/6*(B**3 - B)*(x2 - x1)**2

        y[i] = A*y1 + B*y2 + C*fpp[j] + D*fpp[j+1]
        dAdx = -1.0/(x2 - x1)
        dBdx = -dAdx
        dCdx = 1.0/6*(3*A**2 - 1)*dAdx*(x2 - x1)**2
        dDdx = 1.0/6*(3*B**2 - 1)*dBdx*(x2 - x1)**2
        dydx[i] = dAdx*y1 + dBdx*y2 + dCdx*fpp[j] + dDdx*fpp[j+1]

    if n == 1:
        y = y[0]
        dydx = dydx[0]

    return y


def trapz_deriv(y, x):
    """trapezoidal integration and derivatives with respect to integrand or variable."""

    dI_dy = np.gradient(x)
    dI_dy[0] /= 2
    dI_dy[-1] /= 2

    dI_dx = -np.gradient(y)
    dI_dx[0] = -0.5*(y[0] + y[1])
    dI_dx[-1] = 0.5*(y[-1] + y[-2])

    return dI_dy, dI_dx


def _smooth_maxmin(yd, ymax, maxmin, pct_offset=0.01, dyd=None):

    yd, n = _checkIfFloat(yd)

    y1 = (1-pct_offset)*ymax
    y2 = (1+pct_offset)*ymax

    dy1 = (1-pct_offset)
    dy2 = (1+pct_offset)

    if maxmin == 'min':
        f1 = y1
        f2 = ymax
        g1 = 1.0
        g2 = 0.0
        idx_constant = yd >= y2

        df1 = dy1
        df2 = 1.0


    elif maxmin == 'max':
        f1 = ymax
        f2 = y2
        g1 = 0.0
        g2 = 1.0
        idx_constant = yd <= y1

        df1 = 1.0
        df2 = dy2

    f = CubicSplineSegment(y1, y2, f1, f2, g1, g2)

    # main region
    ya = np.copy(yd)
    if dyd is None:
        dya_dyd = np.ones_like(yd)
    else:
        dya_dyd = np.copy(dyd)

    dya_dymax = np.zeros_like(ya)

    # cubic spline region
    idx = np.logical_and(yd > y1, yd < y2)
    ya[idx] = f.eval(yd[idx])
    dya_dyd[idx] = f.eval_deriv(yd[idx])
    dya_dymax[idx] = f.eval_deriv_params(yd[idx], dy1, dy2, df1, df2, 0.0, 0.0)

    # constant region
    ya[idx_constant] = ymax
    dya_dyd[idx_constant] = 0.0
    dya_dymax[idx_constant] = 1.0

    if n == 1:
        ya = ya[0]
        dya_dyd = dya_dyd[0]
        dya_dymax = dya_dymax[0]


    return ya, dya_dyd, dya_dymax


def smooth_max(yd, ymax, pct_offset=0.01, dyd=None):
    """array max, uses cubic spline to smoothly transition.  derivatives with respect to array and max value.
    width of transition can be controlled, and chain rules for differentiation"""
    return _smooth_maxmin(yd, ymax, 'max', pct_offset, dyd)


def smooth_min(yd, ymin, pct_offset=0.01, dyd=None):
    """array min, uses cubic spline to smoothly transition.  derivatives with respect to array and min value.
    width of transition can be controlled, and chain rules for differentiation"""
    return _smooth_maxmin(yd, ymin, 'min', pct_offset, dyd)



def smooth_abs(x, dx=0.01):
    """smoothed version of absolute vaue function, with quadratic instead of sharp bottom.
    Derivative w.r.t. variable of interest.  Width of quadratic can be controlled"""

    x, n = _checkIfFloat(x)

    y = np.abs(x)
    idx = np.logical_and(x > -dx, x < dx)
    y[idx] = x[idx]**2/(2.0*dx) + dx/2.0

    # gradient
    dydx = np.ones_like(x)
    dydx[x <= -dx] = -1.0
    dydx[idx] = x[idx]/dx


    if n == 1:
        y = y[0]
        dydx = dydx[0]

    return y, dydx



def nodal2sectional(x):
    """Averages nodal data to be length-1 vector of sectional data

    INPUTS:
    ----------
    x   : float vector, nodal data

    OUTPUTS:
    -------
    y   : float vector,  sectional data
    """
    y = 0.5*(x[:-1] + x[1:])
    dy = np.c_[0.5*np.eye(y.size), np.zeros(y.size)]
    dy[np.arange(y.size), 1+np.arange(y.size)] = 0.5
    return y, dy


def sectional2nodal(x):
    return np.r_[x[0], np.convolve(x, [0.5, 0.5], 'valid'), x[-1]]


def cubic_spline_eval(x1, x2, f1, f2, g1, g2, x):

    spline = CubicSplineSegment(x1, x2, f1, f2, g1, g2)
    return spline.eval(x)



class CubicSplineSegment(object):
    """cubic splines and the their derivatives with with respect to the variables and the parameters"""

    def __init__(self, x1, x2, f1, f2, g1, g2):

        self.x1 = x1
        self.x2 = x2

        self.A = np.array([[x1**3, x1**2, x1, 1.0],
                  [x2**3, x2**2, x2, 1.0],
                  [3*x1**2, 2*x1, 1.0, 0.0],
                  [3*x2**2, 2*x2, 1.0, 0.0]])
        self.b = np.array([f1, f2, g1, g2])

        self.coeff = np.linalg.solve(self.A, self.b)

        self.poly = np.polynomial.Polynomial(self.coeff[::-1])


    def eval(self, x):
        return self.poly(x)


    def eval_deriv(self, x):
        polyd = self.poly.deriv()
        return polyd(x)


    def eval_deriv_params(self, xvec, dx1, dx2, df1, df2, dg1, dg2):

        x1 = self.x1
        x2 = self.x2
        dA_dx1 = np.array([[3*x1**2, 2*x1, 1.0, 0.0],
                  [0.0, 0.0, 0.0, 0.0],
                  [6*x1, 2.0, 0.0, 0.0],
                  [0.0, 0.0, 0.0, 0.0]])
        dA_dx2 = np.array([[0.0, 0.0, 0.0, 0.0],
                  [3*x2**2, 2*x2, 1.0, 0.0],
                  [0.0, 0.0, 0.0, 0.0],
                  [6*x2, 2.0, 0.0, 0.0]])
        df = np.array([df1, df2, dg1, dg2])
        c = np.array(self.coeff).T

        n = len(xvec)
        dF = np.zeros(n)
        for i in range(n):
            x = np.array([xvec[i]**3, xvec[i]**2, xvec[i], 1.0])
            d = np.linalg.solve(self.A.T, x)
            dF_dx1 = -d @ dA_dx1 @ c
            dF_dx2 = -d @ dA_dx2 @ c
            dF_df = np.linalg.solve(self.A.T, x)
            dF[i] = np.dot(dF_df, df) + dF_dx1*dx1 + dF_dx2*dx2

        return dF




# def print_vars(comp, list_type='inputs', prefix='', astable=False):
#
#     comp_reserved = ['driver']
#     reserved = ['missing_deriv_policy', 'force_execute', 'directory', 'force_fd', 'printvars', 'exec_count']
#
#     # recursively search for subassemblies
#     if IAssembly.providedBy(comp):  # outer needs to be an assembly
#         subcomp = comp.list_components()
#
#         for subc_name in subcomp:
#             if subc_name in comp_reserved:
#                 continue
#
#             subc = getattr(comp, subc_name)
#             if IAssembly.providedBy(subc):  # inner must be an subassembly
#                 if prefix is not None:
#                     newprefix = prefix + '.' + subc_name
#                 else:
#                     newprefix = subc_name
#                 print_vars(subc, list_type, newprefix, astable)
#                 continue
#
#
#     if list_type == 'inputs':
#         thelist = comp.list_inputs(connected=False)
#     elif list_type == 'outputs':
#         thelist = comp.list_outputs(connected=False)
#     elif list_type == 'vars':
#         thelist = comp.list_vars()
#
#     for name in thelist:
#
#         if name in reserved:
#             continue
#
#         trait = comp.get_trait(name)
#         thetype = trait.trait_type.__str__().split()[0].split('.')[-1]
#
#         if thetype == 'VarTree':
#             vartree = getattr(comp, name)
#             if prefix is not None:
#                 newprefix = prefix + '.' + name
#             else:
#                 newprefix = name
#             print_vars(vartree, 'vars', newprefix, astable)
#             continue
#
#         units = trait.units
#         desc = trait.desc
#         default = trait.default
#         # print trait.category
#
#         if units is None:
#             description = '(' + thetype + ')'
#         else:
#             description = '(' + thetype + ', ' + units + ')'
#
#         if desc is not None:
#             description += ': ' + desc
#
#
#         if prefix is not '':
#             name = prefix + '.' + name
#
#         if not astable:
#             name = name + ' = ' + str(default)
#             print name + '  # ' + description
#
#         else:
#
#             if not units:
#                 units = ''
#             if not desc:
#                 desc = ''
#             strdefault = str(default)
#             if strdefault == '<undefined>':
#                 strdefault = ''
#
#             print '{0:15}\t{1:10}\t{2:15}\t{3:10}\t{4}'.format(name, thetype, strdefault, units, desc)






def _getvar(comp, name):
    vars = name.split('.')
    base = comp
    for var in vars:
        base = getattr(base, var)

    return base


def _setvar(comp, name, value):
    vars = name.split('.')
    base = comp
    for i in range(len(vars)-1):
        base = getattr(base, vars[i])

    setattr(base, vars[-1], value)


def _explodeall(comp, vtype='inputs'):

    if vtype == 'inputs':
        alloutputs = comp.list_inputs()
    else:
        alloutputs = comp.list_outputs()

    maybe_more_vartrees = True

    while maybe_more_vartrees:
        maybe_more_vartrees = False

        for out in alloutputs:
            try:
                vars = getattr(comp, out).list_vars()
                alloutputs.remove(out)
                for var in vars:
                    alloutputs.append(out + '.' + var)
                maybe_more_vartrees = True

            except Exception:
                pass

    return alloutputs


def _getColumnOfOutputs(comp, outputs, m):

    # fill out column of outputs
    m1 = 0
    m2 = 0
    f = np.zeros(m)
    for i, out in enumerate(outputs):

        # get function value at center
        fsub = _getvar(comp, out)
        if np.array(fsub).shape == ():
            lenf = 1
        else:
            fsub = np.copy(fsub)  # so not pointed to same memory address
            lenf = len(fsub)

        m2 += lenf

        f[m1:m2] = fsub

        m1 = m2

    return f


### USING OLD NUMPY SRC FOR PMT-FUNCTION INSTEAD OF SWITCHING TO ANNOYING NUMPY-FINANCIAL
_when_to_num = {'end':0, 'begin':1, 'e':0, 'b':1, 0:0, 1:1, 'beginning':1, 'start':1,'finish':0}

def _convert_when(when):
    #Test to see if when has already been converted to ndarray
    #This will happen if one function calls another, for example ppmt
    if isinstance(when, np.ndarray):
        return when
    try:
        return _when_to_num[when]
    except (KeyError, TypeError):
        return [_when_to_num[x] for x in when]
    
def pmt(rate, nper, pv, fv=0, when='end'):
    when = _convert_when(when)
    (rate, nper, pv, fv, when) = map(np.array, [rate, nper, pv, fv, when])
    temp = (1 + rate)**nper
    mask = (rate == 0)
    masked_rate = np.where(mask, 1, rate)
    fact = np.where(mask != 0, nper,
                    (1 + masked_rate*when)*(temp - 1)/masked_rate)
    return -(fv + pv*temp) / fact

'''
def check_for_missing_unit_tests(modules):
    """A heuristic check to find components that don't have a corresonding unit test
    for its gradients.

    Parameters
    ----------
    modules : list(str)
        a list of modules to check for missin ggradients
    """
    import sys
    import inspect
    import ast

    thisfilemembers = inspect.getmembers(sys.modules['__main__'], lambda member: inspect.isclass(member) and member.__module__ == '__main__')
    tests = [name for name, classname in thisfilemembers]

    totest = []
    tomod = []
    reserved = ['Assembly', 'Slot', 'ImplicitComponent']
    for mod in modules:
        modulemembers = inspect.getmembers(mod, inspect.isclass)
        fname = mod.__file__
        if fname[-1] == 'c':
            fname = fname[:-1]
        f = open(fname, 'r')
        p = ast.parse(f.read())
        f.close()
        nonimported = [node.name for node in ast.walk(p) if isinstance(node, ast.ClassDef)]
        for name, classname in modulemembers:
            bases = classname.__bases__
            if 'Base' not in name and name not in reserved and name in nonimported and len(bases) > 0:
                base = bases[0].__name__
                if base == 'Component' or 'Base' in base:
                    totest.append(name)
                    tomod.append(mod.__name__)

    for mod, test in zip(tomod, totest):
        if 'Test'+test not in tests:
            print('!!! There does not appear to be a unit test for:', mod + '.' + test)


def check_gradient_unit_test(unittest, comp, fd='central', step_size=1e-6, tol=1e-6, display=False,
        show_missing_warnings=True, show_scaling_warnings=False, min_grad=1e-6, max_grad=1e6):
    """compare provided analytic gradients to finite-difference gradients with unit testing.
    Same as check_gradient, but provides a unit test for each gradient for convenience.
    the unit tests checks that the error for each gradient is less than tol.

    Parameters
    ----------
    comp : obj
        An OpenMDAO component that provides analytic gradients through provideJ()
    fd : str
        the type of finite difference to use.  options are central or forward
    step_size : float
        step size to use in finite differencing
    tol : float
        tolerance for how close the gradients should agree to
    display : boolean
        if True, display results for each gradient
    show_missing_warnings: boolean
        if True, warn for gradients that were not provided
        (they may be ones that are unecessary, but may be ones that were accidentally skipped)
    show_scaling_warnings: boolean
        if True, warn for gradients that are either very small or very large which may lead
        to challenges in solving the full linear system
    min_grad/max_grad : float
        quantifies what "very small" or "very large" means when using show_scaling_warnings
    """

    names, errors = check_gradient(comp, fd, step_size, tol, display, show_missing_warnings,
        show_scaling_warnings, min_grad, max_grad)

    for name, err in zip(names, errors):
        try:
            unittest.assertLessEqual(err, tol)
        except AssertionError(e):
            print('*** error in:', name)
            raise e


def check_gradient(comp, fd='central', step_size=1e-6, tol=1e-6, display=False,
        show_missing_warnings=True, show_scaling_warnings=False, min_grad=1e-6, max_grad=1e6):
    """compare provided analytic gradients to finite-difference gradients

    Parameters
    ----------
    comp : obj
        An OpenMDAO component that provides analytic gradients through provideJ()
    fd : str
        the type of finite difference to use.  options are central or forward
    step_size : float
        step size to use in finite differencing
    tol : float
        tolerance for how close the gradients should agree to
    display : boolean
        if True, display results for each gradient
    show_missing_warnings: boolean
        if True, warn for gradients that were not provided
        (they may be ones that are unecessary, but may be ones that were accidentally skipped)
    show_scaling_warnings: boolean
        if True, warn for gradients that are either very small or very large which may lead
        to challenges in solving the full linear system
    min_grad/max_grad : float
        quantifies what "very small" or "very large" means when using show_scaling_warnings

    Returns
    -------
    names : array(str)
        list of the names of all the gradients
    errorvec : array(float)
        list of all the errors for the gradients.  If the magnitude of the gradient is less than
        tol, then an absolute error is used, otherwise a relative error is used.

    """

    inputs, outputs = comp.list_deriv_vars()



    if show_missing_warnings:
        all_inputs = _explodeall(comp, vtype='inputs')
        all_outputs = _explodeall(comp, vtype='outputs')
        reserved_inputs = ['missing_deriv_policy', 'directory', 'force_fd', 'force_execute', 'eval_only']
        reserved_outputs = ['derivative_exec_count', 'itername', 'exec_count']
        potential_missed_inputs = list(set(all_inputs) - set(reserved_inputs) - set(inputs))
        potential_missed_outputs = list(set(all_outputs) - set(reserved_outputs) - set(outputs))

        if len(potential_missed_inputs) > 0 or len(potential_missed_outputs) > 0:
            print()
            print('*** Warning: ' + comp.__class__.__name__ + ' does not supply derivatives for the following')
            print('\tinputs:', potential_missed_inputs)
            print('\toutputs:', potential_missed_outputs)
            print()


    comp.run()
    J = comp.provideJ()

    # compute size of Jacobian
    m = 0
    mvec = []  # size of each output
    cmvec = []  # cumulative size of outputs
    nvec = []  # size of each input
    cnvec = []  # cumulative size of inputs
    for out in outputs:
        f = _getvar(comp, out)
        if np.array(f).shape == ():
            msub = 1
        else:
            msub = len(f)
        m += msub
        mvec.append(msub)
        cmvec.append(m)
    n = 0
    for inp in inputs:
        x = _getvar(comp, inp)
        if np.array(x).shape == ():
            nsub = 1
        else:
            nsub = len(x)
        n += nsub
        nvec.append(nsub)
        cnvec.append(n)

    JFD = np.zeros((m, n))

    if J.shape != JFD.shape:
        raise TypeError('Incorrect Jacobian size. Your provided Jacobian is of shape {}, but it should be ({}, {})'.format(J.shape, m, n))


    # fill out column of outputs
    f = _getColumnOfOutputs(comp, outputs, m)

    n1 = 0

    for j, inp in enumerate(inputs):

        # get x value at center (save location)
        x = _getvar(comp, inp)
        if np.array(x).shape == ():
            x0 = x
            lenx = 1
        else:
            x = np.copy(x)  # so not pointing to same memory address
            x0 = np.copy(x)
            lenx = len(x)

        for k in range(lenx):

            # take a step
            if lenx == 1:
                h = np.abs(step_size*x)
                if h < step_size:
                    h = step_size
                x += h
            else:
                h = np.abs(step_size*x[k])
                if h < step_size:
                    h = step_size
                x[k] += h
            _setvar(comp, inp, x)
            comp.run()

            # fd
            fp = _getColumnOfOutputs(comp, outputs, m)

            if fd == 'central':

                # step back
                if lenx == 1:
                    x -= 2*h
                else:
                    x[k] -= 2*h
                _setvar(comp, inp, x)
                comp.run()

                fm = _getColumnOfOutputs(comp, outputs, m)

                deriv = (fp - fm)/(2*h)

            else:
                deriv = (fp - f)/h


            JFD[:, n1+k] = deriv

            # reset state
            x = np.copy(x0)
            _setvar(comp, inp, x0)
            comp.run()

        n1 += lenx


    # error checking
    namevec = []
    errorvec = []

    if display:
        print('{:<20} ({}) {:<10} ({}, {})'.format('error', 'errortype', 'name', 'analytic', 'fd'))
        print()

    for i in range(m):
        for j in range(n):

            # get corresonding variables names
            for ii in range(len(mvec)):
                if cmvec[ii] > i:
                    oname = 'd_' + outputs[ii]

                    if mvec[ii] > 1:  # need to print indices
                        subtract = 0
                        if ii > 0:
                            subtract = cmvec[ii-1]
                        idx = i - subtract
                        oname += '[' + str(idx) + ']'

                    break
            for jj in range(len(nvec)):
                if cnvec[jj] > j:
                    iname = 'd_' + inputs[jj]

                    if nvec[jj] > 1:  # need to print indices
                        subtract = 0
                        if jj > 0:
                            subtract = cnvec[jj-1]
                        idx = j - subtract
                        iname += '[' + str(idx) + ']'

                    break
            name = oname + ' / ' + iname

            # compute error
            if np.abs(J[i, j]) <= tol:
                errortype = 'absolute'
                error = J[i, j] - JFD[i, j]
            else:
                errortype = 'relative'
                error = 1.0 - JFD[i, j]/J[i, j]
            error = np.abs(error)

            # display
            if error > tol:
                star = ' ***** '
            else:
                star = ''

            if display:
                output = '{}{:<20} ({}) {}: ({}, {})'.format(star, error, errortype, name, J[i, j], JFD[i, j])
                print(output)

            if show_scaling_warnings and J[i, j] != 0 and np.abs(J[i, j]) < min_grad:
                print('*** Warning: The following analytic gradient is very small and may need to be scaled:')
                print('\t(' + comp.__class__.__name__ + ') ' + name + ':', J[i, j])

            if show_scaling_warnings and np.abs(J[i, j]) > max_grad:
                print('*** Warning: The following analytic gradient is very large and may need to be scaled:')
                print('\t(' + comp.__class__.__name__ + ') ' + name + ':', J[i, j])


            # save
            namevec.append(name)
            errorvec.append(error)

    return namevec, errorvec
'''



# if __name__ == '__main__':



    # xpt = np.array([1.0, 2.0, 4.0, 6.0, 10.0, 12.0])
    # ypt = np.array([5.0, 12.0, 14.0, 16.0, 21.0, 29.0])

    # # interpolate  (extrapolation will work, but beware the results may be silly)
    # n = 50
    # x = np.linspace(0.0, 13.0, n)
    # y = cubic_with_deriv(x, xpt, ypt)

    # import matplotlib.pyplot as plt
    # plt.plot(xpt, ypt, 'o')
    # plt.plot(x, y, '-')
    # plt.show()
